import typing

from sklearn.metrics import (
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
)

from mlfoundry.enums import ModelType
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger
from mlfoundry.metrics.v2.base_metrics import BaseMetrics


class RegressionMetrics(BaseMetrics):
    MEAN_SQUARED_ERROR = "mean_squared_error"
    MEAN_ABSOLUTE_ERROR = "mean_absolute_error"
    R2_SCORE = "r2_score"
    ROOT_MEAN_SQUARE_ERROR = "root_mean_squared_error"
    EXPLAINED_VARIANCE_SCORE = "explained_variance_score"

    @property
    def model_type(self) -> ModelType:
        return ModelType.REGRESSION

    def get_metric_name_to_function_map(self) -> typing.Dict[str, typing.Callable]:
        return {
            self.MEAN_SQUARED_ERROR: self.compute_mean_squared_error,
            self.MEAN_ABSOLUTE_ERROR: self.compute_mean_absolute_error,
            self.R2_SCORE: self.compute_r2_score,
            self.ROOT_MEAN_SQUARE_ERROR: self.compute_root_mean_squared_error,
            self.EXPLAINED_VARIANCE_SCORE: self.compute_explained_variance_score,
        }

    def validate(self, predictions, actuals, **kwargs):
        if predictions is None:
            raise MlFoundryException("predictions cannot be None")
        if actuals is None:
            raise MlFoundryException("actuals cannot be none")

    @staticmethod
    def compute_mean_squared_error(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return mean_squared_error(actuals, predictions)

    @staticmethod
    def compute_mean_absolute_error(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return mean_absolute_error(actuals, predictions)

    @staticmethod
    def compute_r2_score(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return r2_score(actuals, predictions)

    @staticmethod
    def compute_root_mean_squared_error(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return mean_squared_error(actuals, predictions, squared=False)

    @staticmethod
    def compute_explained_variance_score(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return explained_variance_score(actuals, predictions)



import os
import posixpath
import tempfile
import time
import typing
import uuid

import pandas as pd
from mlflow.entities import RunLog
from mlflow.tracking import MlflowClient
from whylogs import DatasetProfile

from mlfoundry import log_types
from mlfoundry.constants import RUN_LOGS_DIR
from mlfoundry.dataset.serde_utils import (
    load_dataframe,
    load_series,
    save_dataframe,
    save_series,
)
from mlfoundry.dataset.types import DatasetSchema, DatasetStats, Profiles
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger

DATASET_LOG_DIR = posixpath.join(RUN_LOGS_DIR, "dataset")


class TabularDatasetSerDe:
    _FEATURE_ENTITY_NAME = "features"
    _PREDICTIONS_ENTITY_NAME = "predictions"
    _ACTUALS_ENTITY_NAME = "actuals"
    _SCHEMA_FILE_NAME = "schema.json"
    _STATS_FILE_NAME = "stats.json"
    _FEATURES_PROFILE_FILE_NAME = "features_profile.pb.bin"
    _PREDICTIONS_PROFILE_FILE_NAME = "predictions_profile.pb.bin"
    _ACTUALS_PROFILE_FILE_NAME = "actuals_profile.pb.bin"

    def __init__(self, mlflow_client: MlflowClient):
        self.mlflow_client: MlflowClient = mlflow_client
        self.dataset_prefix = DATASET_LOG_DIR

    def _load_file(
        self,
        run_id: str,
        artifact_path: str,
        deserializer: typing.Callable,
    ):
        with tempfile.TemporaryDirectory() as local_dir:
            local_path = self.mlflow_client.download_artifacts(
                run_id, artifact_path, local_dir
            )
            return deserializer(local_path)

    def _get_run_log(
        self, run_id: str, dataset_name: str, log_type: str
    ) -> typing.Optional[RunLog]:
        run_logs = self.mlflow_client.list_run_logs(
            run_uuid=run_id, key=dataset_name, log_type=log_type
        )
        if len(run_logs) == 0:
            return None
        return run_logs[0]

    def validate_dataset_is_not_already_logged(self, run_id, dataset_name):
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifact.get_log_type(),
        )
        if run_log is None:
            return
        raise MlFoundryException(f"dataset {dataset_name} is already logged")

    def save_dataset(
        self,
        run_id: str,
        dataset_name: str,
        features: pd.DataFrame,
        predictions: typing.Optional[pd.Series],
        actuals: typing.Optional[pd.Series],
        schema: DatasetSchema,
        stats: DatasetStats,
        profiles: Profiles,
        only_stats: bool,
    ) -> str:
        artifact_path = posixpath.join(
            self.dataset_prefix, dataset_name, str(uuid.uuid4())
        )
        timestamp = int(time.time() * 1000)
        run_logs = [
            log_types.DatasetArtifact(artifact_path=artifact_path).to_run_log(
                key=dataset_name, timestamp=timestamp
            )
        ]
        with tempfile.TemporaryDirectory() as local_dir:
            if not only_stats:
                file_name, file_format = save_dataframe(
                    local_dir=local_dir,
                    entity_name=TabularDatasetSerDe._FEATURE_ENTITY_NAME,
                    dataframe=features,
                )
                run_logs.append(
                    log_types.DatasetArtifactFeatures(
                        artifact_path=posixpath.join(artifact_path, file_name),
                        format=file_format,
                    ).to_run_log(key=dataset_name, timestamp=timestamp)
                )

                if predictions is not None:
                    file_name, file_format = save_series(
                        local_dir=local_dir,
                        entity_name=TabularDatasetSerDe._PREDICTIONS_ENTITY_NAME,
                        series=predictions,
                    )
                    run_logs.append(
                        log_types.DatasetArtifactPredictions(
                            artifact_path=posixpath.join(artifact_path, file_name),
                            format=file_format,
                        ).to_run_log(key=dataset_name, timestamp=timestamp)
                    )
                if actuals is not None:
                    file_name, file_format = save_series(
                        local_dir=local_dir,
                        entity_name=TabularDatasetSerDe._ACTUALS_ENTITY_NAME,
                        series=actuals,
                    )
                    run_logs.append(
                        log_types.DatasetArtifactActuals(
                            artifact_path=posixpath.join(artifact_path, file_name),
                            format=file_format,
                        ).to_run_log(key=dataset_name, timestamp=timestamp)
                    )

            run_logs.append(
                schema.to_log().to_run_log_as_artifact(
                    key=dataset_name,
                    run_id=run_id,
                    mlflow_client=self.mlflow_client,
                    file_name=TabularDatasetSerDe._SCHEMA_FILE_NAME,
                    artifact_path=artifact_path,
                )
            )
            run_logs.append(
                stats.to_log().to_run_log_as_artifact(
                    key=dataset_name,
                    run_id=run_id,
                    mlflow_client=self.mlflow_client,
                    file_name=TabularDatasetSerDe._STATS_FILE_NAME,
                    artifact_path=artifact_path,
                )
            )

            features_profile_path = os.path.join(
                local_dir, TabularDatasetSerDe._FEATURES_PROFILE_FILE_NAME
            )
            profiles.features.write_protobuf(features_profile_path)
            run_logs.append(
                log_types.DatasetArtifactFeaturesProfile(
                    artifact_path=posixpath.join(
                        artifact_path, TabularDatasetSerDe._FEATURES_PROFILE_FILE_NAME
                    )
                ).to_run_log(key=dataset_name, timestamp=timestamp)
            )

            if profiles.predictions is not None:
                predictions_profile_path = os.path.join(
                    local_dir, TabularDatasetSerDe._PREDICTIONS_PROFILE_FILE_NAME
                )
                profiles.predictions.write_protobuf(predictions_profile_path)
                run_logs.append(
                    log_types.DatasetArtifactPredictionsProfile(
                        artifact_path=posixpath.join(
                            artifact_path,
                            TabularDatasetSerDe._PREDICTIONS_PROFILE_FILE_NAME,
                        )
                    ).to_run_log(key=dataset_name, timestamp=timestamp)
                )

            if profiles.actuals is not None:
                actuals_profile_path = os.path.join(
                    local_dir, TabularDatasetSerDe._ACTUALS_PROFILE_FILE_NAME
                )
                profiles.actuals.write_protobuf(actuals_profile_path)
                run_logs.append(
                    log_types.DatasetArtifactActualsProfile(
                        artifact_path=posixpath.join(
                            artifact_path,
                            TabularDatasetSerDe._ACTUALS_PROFILE_FILE_NAME,
                        )
                    ).to_run_log(key=dataset_name, timestamp=timestamp)
                )

            self.mlflow_client.log_artifacts(
                run_id=run_id, local_dir=local_dir, artifact_path=artifact_path
            )
            self.mlflow_client.insert_run_logs(run_uuid=run_id, run_logs=run_logs)

    def load_schema(self, run_id: str, dataset_name: str) -> DatasetSchema:
        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=run_id,
            key=dataset_name,
            log_type=log_types.DatasetSchemaRunLog.get_log_type(),
        )
        return DatasetSchema.parse_obj(
            log_types.DatasetSchemaRunLog.from_run_log(run_log=run_log).value
        )

    def load_stats(self, run_id: str, dataset_name: str) -> DatasetStats:
        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=run_id,
            key=dataset_name,
            log_type=log_types.DatasetStatsRunLog.get_log_type(),
        )
        return DatasetStats.parse_obj(
            log_types.DatasetStatsRunLog.from_run_log(run_log=run_log).value
        )

    def _load_features_profile(self, run_id: str, dataset_name: str):
        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=run_id,
            key=dataset_name,
            log_type=log_types.DatasetArtifactFeaturesProfile.get_log_type(),
        )
        artifact_path = log_types.DatasetArtifactFeaturesProfile.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id,
            artifact_path=artifact_path,
            deserializer=DatasetProfile.read_protobuf,
        )

    def _load_predictions_profile(self, run_id: str, dataset_name: str):
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactPredictionsProfile.get_log_type(),
        )
        if run_log is None:
            logger.info(f"prediction profile for {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactPredictionsProfile.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id,
            artifact_path=artifact_path,
            deserializer=DatasetProfile.read_protobuf,
        )

    def _load_actuals_profile(self, run_id: str, dataset_name: str):
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactActualsProfile.get_log_type(),
        )
        if run_log is None:
            logger.info(f"actuals profile for {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactActualsProfile.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id,
            artifact_path=artifact_path,
            deserializer=DatasetProfile.read_protobuf,
        )

    def load_profiles(self, run_id: str, dataset_name: str) -> Profiles:
        features_profile = self._load_features_profile(run_id, dataset_name)
        actuals_profile = self._load_actuals_profile(run_id, dataset_name)
        predictions_profile = self._load_predictions_profile(run_id, dataset_name)
        return Profiles(
            features=features_profile,
            actuals=actuals_profile,
            predictions=predictions_profile,
        )

    def load_features(
        self, run_id: str, dataset_name: str
    ) -> typing.Optional[pd.DataFrame]:
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactFeatures.get_log_type(),
        )
        if run_log is None:
            logger.info(f"feature of dataset {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactFeatures.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id, artifact_path=artifact_path, deserializer=load_dataframe
        )

    def load_actuals(
        self, run_id: str, dataset_name: str
    ) -> typing.Optional[pd.DataFrame]:
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactActuals.get_log_type(),
        )
        if run_log is None:
            logger.info(f"actuals of dataset {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactActuals.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id, artifact_path=artifact_path, deserializer=load_series
        )

    def load_predictions(
        self, run_id: str, dataset_name: str
    ) -> typing.Optional[pd.DataFrame]:
        run_log = self._get_run_log(
            run_id=run_id,
            dataset_name=dataset_name,
            log_type=log_types.DatasetArtifactPredictions.get_log_type(),
        )
        if run_log is None:
            logger.info(f"actuals of dataset {dataset_name} not found")
            return None

        artifact_path = log_types.DatasetArtifactPredictions.from_run_log(
            run_log
        ).artifact_path
        return self._load_file(
            run_id=run_id, artifact_path=artifact_path, deserializer=load_series
        )



import datetime
import os
import platform
import queue
import re
import time
import weakref
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import urlencode, urljoin, urlsplit, urlunsplit

import mlflow
import pandas as pd
import whylogs
from mlflow.entities import (
    CustomMetric,
    Metric,
    ModelSchema,
    Param,
    RunInfo,
    RunStatus,
    RunTag,
)
from mlflow.tracking import MlflowClient

from mlfoundry import amplitude, constants, enums, version
from mlfoundry.dataset import DataSet, TabularDatasetDriver
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.git_info import GitInfo
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.log_types import Image, Plot
from mlfoundry.log_types.artifacts.artifact import ArtifactPath, ArtifactVersion
from mlfoundry.log_types.artifacts.general_artifact import _log_artifact_version
from mlfoundry.log_types.artifacts.model import ModelVersion, _log_model_version
from mlfoundry.logger import logger
from mlfoundry.metrics.v2 import ComputedMetrics
from mlfoundry.metrics.v2 import get_metrics_calculator as get_metrics_calculator_v2
from mlfoundry.run_utils import (
    ParamsType,
    flatten_dict,
    log_artifact_blob,
    process_params,
)
from mlfoundry.session import ACTIVE_RUNS


class MlFoundryRun:
    """MlFoundryRun."""

    # TODO (nikunjbjj): Seems like these artifact locations are defined in constants.py for loading.
    #  Need to change this.
    S3_DATASET_PATH = "datasets"
    S3_STATS_PATH = "stats"
    S3_WHYLOGS_PATH = "whylogs"
    S3_METRICS_PATH = "multi_dimensional_metrics"

    VALID_PARAM_AND_METRIC_NAMES = re.compile(r"^[A-Za-z0-9_\-\. /]+$")

    def __init__(
        self,
        experiment_id: str,
        run_id: str,
        **kwargs,
    ):
        """__init__.

        Args:
            experiment_id (str): experiment_id
            run_id (str): run_id
        """
        # TODO (chiragjn): rename experiment to project everywhere
        self._experiment_id = str(experiment_id)
        # TODO (chiragjn): mlflow_client be a protected/private member
        self.mlflow_client = MlflowClient()
        self._project_name = self.mlflow_client.get_experiment(self._experiment_id).name

        self._run_id = run_id
        self._run_info: Optional[RunInfo] = None

        self._dataset_module: TabularDatasetDriver = TabularDatasetDriver(
            mlflow_client=self.mlflow_client, run_id=run_id
        )

        self._terminate_called = False
        ACTIVE_RUNS.add_run(self)
        print(f"Link to the dashboard for the run: {self.dashboard_link}")

    def _get_run_info(self) -> RunInfo:
        if self._run_info is not None:
            return self._run_info

        self._run_info = self.mlflow_client.get_run(self.run_id).info
        return self._run_info

    @property
    def run_id(self) -> str:
        return self._run_id

    @property
    def run_name(self) -> str:
        return self._get_run_info().name

    @property
    def fqn(self) -> str:
        return self._get_run_info().fqn

    @property
    def status(self) -> str:
        return self.mlflow_client.get_run(self.run_id).info.status

    @property
    def project_name(self) -> str:
        return self._project_name

    def __repr__(self) -> str:
        return f"<{type(self).__name__} at 0x{id(self):x}: run={self.fqn!r}>"

    def __enter__(self):
        return self

    def _terminate_run_if_running(self, termination_status: RunStatus):
        """_terminate_run_if_running.

        Args:
            termination_status (RunStatus): termination_status
        """
        if self._terminate_called:
            return

        # Prevent double execution for termination
        self._terminate_called = True
        ACTIVE_RUNS.remove_run(self)

        current_status = self.status
        termination_status = RunStatus.to_string(termination_status)
        try:
            # we do not need to set any termination status unless the run was in RUNNING state
            if current_status != RunStatus.to_string(RunStatus.RUNNING):
                return
            logger.info("Setting run status of %r to %r", self.fqn, termination_status)
            self.mlflow_client.set_terminated(self.run_id, termination_status)
        except Exception as e:
            logger.warning(
                f"failed to set termination status {termination_status} due to {e}"
            )
        print(f"Finished run: {self.fqn!r}, Dashboard: {self.dashboard_link}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
        self._terminate_run_if_running(status)

    def __del__(self):
        # TODO (chiragjn): Should this be marked as FINISHED or KILLED?
        self._terminate_run_if_running(RunStatus.FINISHED)

    @property
    def dashboard_link(self) -> str:
        """Get Mlfoundry dashboard link for a `run`"""

        base_url = "{uri.scheme}://{uri.netloc}/".format(
            uri=urlsplit(mlflow.get_tracking_uri())
        )

        return urljoin(base_url, f"mlfoundry/{self._experiment_id}/run/{self.run_id}/")

    def end(self, status: RunStatus = RunStatus.FINISHED):
        """End a `run`.

        This function marks the run as `FINISHED`.

        Example:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )
        # ...
        # Model training code
        # ...
        run.end()
        ```

        In case the run was created using the context manager approach,
        We do not need to call this function.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        with client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        ) as run:
            # ...
            # Model training code
            ...
        # `run` will be automatically marked as `FINISHED` or `FAILED`.
        ```
        """
        self._terminate_run_if_running(status)

    def _add_git_info(self, root_path: Optional[str] = None):
        """_add_git_info.

        Args:
            root_path (Optional[str]): root_path
        """
        root_path = root_path or os.getcwd()
        try:
            git_info = GitInfo(root_path)
            tags = [
                RunTag(
                    key=constants.GIT_COMMIT_TAG_NAME,
                    value=git_info.current_commit_sha,
                ),
                RunTag(
                    key=constants.GIT_BRANCH_TAG_NAME,
                    value=git_info.current_branch_name,
                ),
                RunTag(key=constants.GIT_DIRTY_TAG_NAME, value=str(git_info.is_dirty)),
            ]
            remote_url = git_info.remote_url
            if remote_url is not None:
                tags.append(RunTag(key=constants.GIT_REMOTE_URL_NAME, value=remote_url))
            self.mlflow_client.log_batch(run_id=self.run_id, tags=tags)
            log_artifact_blob(
                mlflow_client=self.mlflow_client,
                run_id=self.run_id,
                blob=git_info.diff_patch,
                file_name=constants.PATCH_FILE_NAME,
                artifact_path=constants.PATCH_FILE_ARTIFACT_DIR,
            )
        except Exception as ex:
            # no-blocking
            logger.warning(f"failed to log git info because {ex}")

    def _add_python_mlf_version(self):
        python_version = platform.python_version()
        mlfoundry_version = version.__version__

        tags = [
            RunTag(
                key=constants.PYTHON_VERSION_TAG_NAME,
                value=python_version,
            ),
        ]

        if mlfoundry_version:
            tags.append(
                RunTag(
                    key=constants.MLFOUNDRY_VERSION_TAG_NAME,
                    value=mlfoundry_version,
                )
            )
        else:
            logger.warning("Failed to get MLFoundry version.")

        self.mlflow_client.log_batch(run_id=self.run_id, tags=tags)

    def download_artifact_depricated(
        self, path: str, dest_path: Optional[str] = None
    ) -> str:
        """Downloads a logged `artifact` associated with the current `run`.

        Args:
            path (str): Source artifact path.
            dest_path (Optional[str], optional): Absolute path of the local destination
                directory. If a directory path is passed, the directory must already exist.
                If not passed, the artifact will be downloaded to a newly created directory
                in the local filesystem.

        Returns:
            str: Path of the directory where the artifact is downloaded.

        Examples:
        ```python
        import os
        import mlfoundry

        with open("artifact.txt", "w") as f:
            f.write("hello-world")

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )

        run.log_artifact(local_path="artifact.txt", artifact_path="my-artifacts")

        local_path = run.download_artifact(path="my-artifacts")
        print(f"Artifacts: {os.listdir(local_path)}")

        run.end()
        ```
        """
        if dest_path is None:
            return self.mlflow_client.download_artifacts(self.run_id, path=path)
        elif os.path.isdir(dest_path):
            return self.mlflow_client.download_artifacts(
                self.run_id, path=path, dst_path=dest_path
            )
        else:
            raise MlFoundryException(
                f"Destination path {dest_path} should be an existing directory."
            )

    def log_artifact_depricated(
        self, local_path: str, artifact_path: Optional[str] = None
    ):
        """Logs an artifact for the current `run`.

        An `artifact` is a local file or directory. This function stores the `artifact`
        at the remote artifact storage.

        Args:
            local_path (str): Local path of the file or directory.
            artifact_path (Optional[str], optional): Relative destination path where
                the `artifact` will be stored. If not passed, the `artifact` is stored
                at the root path of the `run`'s artifact storage. The passed
                `artifact_path` should not start with `mlf/`, as `mlf/` directory is
                reserved for `mlfoundry`.

        Examples:
        ```python
        import os
        import mlfoundry

        with open("artifact.txt", "w") as f:
            f.write("hello-world")

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )

        run.log_artifact(local_path="artifact.txt", artifact_path="my-artifacts")

        run.end()
        ```
        """
        # TODO (chiragjn): this api is a little bit confusing, artifact_path is always considered to be a directory
        # which means passing local_path="a/b/c/d.txt", artifact_path="x/y/z.txt" will result in x/y/z.txt/d.txt
        if artifact_path is not None:
            NAMESPACE.validate_namespace_not_used(path=artifact_path)
        if os.path.isfile(local_path):
            _to = os.path.join(artifact_path or "", os.path.basename(local_path))
            logger.info(
                f"Logging {local_path!r} file as artifact to {_to!r}, this might take a while ..."
            )
            self.mlflow_client.log_artifact(
                self.run_id, local_path=local_path, artifact_path=artifact_path
            )
        elif os.path.isdir(local_path):
            _to = artifact_path or "/"
            logger.info(
                f"Logging contents of the {local_path!r} directory as artifacts to {_to!r}, this might take a while ..."
            )
            self.mlflow_client.log_artifacts(
                self.run_id, local_dir=local_path, artifact_path=artifact_path
            )
        else:
            raise MlFoundryException(
                f"local path {local_path} should be an existing file or directory"
            )

    def log_artifact(
        self,
        name: str,
        artifact_paths: List[
            Union[Tuple[str], Tuple[str, Optional[str]], ArtifactPath]
        ],
        description: Optional[str] = None,
        metadata: Optional[Dict] = None,
        step: Optional[int] = 0,
    ) -> ArtifactVersion:
        """Logs an artifact for the current `project`.

        An `artifact` is a list of local files and directories. This function packs the mentioned files and directories in `artifact_paths` and uploads them to remote storage linked to the experiment

        Args:
            name (str): Name of the Artifact. If an artifact with this name already exists under the current project,
                the logged artifact will be added as a new version under that `name`. If no artifact exist with the given
                `name`, the given artifact will be logged as version 1.
            artifact_paths (List[mlfoundry.ArtifactPath], optional): A list of pairs
                of (source path, destination path) to add files and folders
                to the artifact version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the artifact contents to upload to.

                E.g. >>> run.log_artifact(
                     ...     name="xyz",
                     ...     artifact_paths=[
                                mlfoundry.ArtifactPath("foo.txt", "foo/bar/foo.txt"),
                                mlfoundry.ArtifactPath("tokenizer/", "foo/tokenizer/"),
                                mlfoundry.ArtifactPath('bar.text'),
                                ('bar.txt', ),
                                ('foo.txt', 'a/foo.txt')
                             ]
                     ... )
                would result in
                .
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            step (int): step/iteration at which the vesion is being logged, defaults to 0.

        Returns:
            mlfoundry.ArtifactVersion: an instance of `ArtifactVersion` that can be used to download the files,
            or update attributes like description, metadata.

        Examples:
        ```python
        import os
        import mlfoundry

        with open("artifact.txt", "w") as f:
            f.write("hello-world")

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )

        run.log_artifact(
            name="hello-world-file",
            artifact_paths=[mlfoundry.ArtifactPath('artifact.txt', 'a/b/')]
        )

        run.end()
        ```
        """
        if not artifact_paths:
            raise MlFoundryException(
                "artifact_paths cannot be empty, atleast one artifact_path must be passed"
            )

        return _log_artifact_version(
            self,
            name=name,
            artifact_paths=artifact_paths,
            description=description,
            metadata=metadata,
            step=step,
        )

    def log_dataset(
        self,
        dataset_name: str,
        features,
        predictions=None,
        actuals=None,
        only_stats: bool = False,
    ):
        """Log a tabular dataset for the current `run`.

        Log a dataset associated with a run. A dataset is a collection of features,
        predictions and actuals. Datasets are uniquely identified by the `dataset_name`
        under a run. They are immutable, once successfully logged, overwriting it is not allowed.
        Mixed types are not allowed in features, actuals and predictions. However, there can be
        missing data in the form of None, NaN, NA.

        The statistics can be visualized in the mlfoundry dashboard.

        Args:
            dataset_name (str): Name of the dataset. Dataset name should only contain
                letters(a-b, A-B), numbers (0-9), underscores (_) and hyphens (-).
            features: Features associated with this dataset.
                This should be either pandas DataFrame or should be of a
                data type which can be converted to a DataFrame.
            predictions (Iterable,optional): Predictions associated with this dataset and run. This
                should be either pandas Series or should be of a data type which
                can be converted to a Series. This is an optional argument.
            actuals (Iterable, optional): Actuals associated with this dataset and run. This
                should be either pandas Series or should be of a data type which
                can be converted to a Series. This is an optional argument.
            only_stats (bool, optional): If True, then the dataset
                (features, predictions, actuals) is not saved. Only statistics and
                the dataset schema will be persisted. Default is False.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )

        features = [
            {"feature_1": 1, "feature_2": 1.2, "feature_3": "high"},
            {"feature_1": 2, "feature_2": 3.5, "feature_3": "medium"},
        ]
        run.log_dataset(
            dataset_name="train",
            features=features,
            actuals=[1.2, 1.3],
            predictions=[3.1, 4.5],
        )

        run.end()
        ```
        """
        amplitude.track(amplitude.Event.LOG_DATASET)
        logger.info("Logging Dataset, this might take a while ...")
        self._dataset_module.log_dataset(
            dataset_name=dataset_name,
            features=features,
            predictions=predictions,
            actuals=actuals,
            only_stats=only_stats,
        )
        logger.info("Dataset logged successfully")
        split = urlsplit(self.dashboard_link)
        dashboard_link = urlunsplit(
            (
                split.scheme,
                split.netloc,
                split.path,
                urlencode({"tab": "data-feature-metrics"}),
                None,
            )
        )
        print(f"To visualize the logged dataset, click on the link {dashboard_link}")

    def get_dataset(self, dataset_name: str) -> Optional[DataSet]:
        """Get a logged dataset by `dataset_name`.

        Args:
            dataset_name (str): Name of the dataset.

        Returns:
            Optional[DataSet]: Returns logged dataset as an instance of
                the `DataSet` class. If the dataset was not logged or
                `only_stats` was set to `True` while logging the dataset,
                the function will return `None`.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )

        features = [
            {"feature_1": 1, "feature_2": 1.2, "feature_3": "high"},
            {"feature_1": 2, "feature_2": 3.5, "feature_3": "medium"},
        ]
        run.log_dataset(
            dataset_name="train",
            features=features,
            actuals=[1.2, 1.3],
            predictions=[3.1, 4.5],
        )
        dataset = run.get_dataset("train")
        print(dataset.features) # This will be in Pandas DataFrame type.
        print(dataset.predictions) # This will be in Pandas Series type.
        print(dataset.actuals) # This will be in Pandas Series type.
        run.end()
        ```
        """
        amplitude.track(amplitude.Event.LOG_DATASET)
        return self._dataset_module.get_dataset(dataset_name=dataset_name)

    def log_metrics(self, metric_dict: Dict[str, Union[int, float]], step: int = 0):
        """Log metrics for the current `run`.

        A metric is defined by a metric name (such as "training-loss") and a
        floating point or integral value (such as `1.2`). A metric is associated
        with a `step` which is the training iteration at which the metric was
        calculated.

        Args:
            metric_dict (Dict[str, Union[int, float]]): A metric name to metric value map.
                metric value should be either `float` or `int`. This should be
                a non-empty dictionary.
            step (int, optional): Training step/iteration at which the metrics
                present in `metric_dict` were calculated. If not passed, `0` is
                set as the `step`.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        run.log_metrics(metric_dict={"accuracy": 0.7, "loss": 0.6}, step=0)
        run.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4}, step=1)

        run.end()
        ```
        """
        # not sure about amplitude tracking here.
        # as the user can use this function in training loop
        # amplitude.track(amplitude.Event.LOG_METRICS)

        try:
            # mlfow_client doesn't have log_metrics api, so we have to use log_batch,
            # This is what internally used by mlflow.log_metrics
            timestamp = int(time.time() * 1000)
            metrics_arr = []
            for key in metric_dict.keys():
                if isinstance(metric_dict[key], str):
                    logger.warning(
                        f"Cannot log metric with string value. Discarding metric {key}={metric_dict[key]}"
                    )
                    continue
                if not self.VALID_PARAM_AND_METRIC_NAMES.match(key):
                    logger.warning(
                        f"Invalid metric name: {key}. Names may only contain alphanumerics, underscores (_), dashes (-), periods (.), spaces ( ), and slashes (/). "
                        f"Discarding metric {key}={metric_dict[key]}"
                    )
                    continue
                metrics_arr.append(Metric(key, metric_dict[key], timestamp, step=step))

            if len(metrics_arr) == 0:
                logger.warning("Cannot log empty metrics dictionary")
                return

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=metrics_arr, params=[], tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        logger.info("Metrics logged successfully")

    def log_params(self, param_dict: ParamsType, flatten_params: bool = False):
        """Logs parameters for the run.

        Parameters or Hyperparameters can be thought of as configurations for a run.
        For example, the type of kernel used in a SVM model is a parameter.
        A Parameter is defined by a name and a string value. Parameters are
        also immutable, we cannot overwrite parameter value for a parameter
        name.

        Args:
            param_dict (ParamsType): A parameter name to parameter value map.
                Parameter values are converted to `str`.

        Examples:
        ### Logging parameters using a `dict`.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        run.log_params({"learning_rate": 0.01, "epochs": 10})

        run.end()
        ```

        ### Logging parameters using `argparse` Namespace object
        ```python
        import argparse
        import mlfoundry

        parser = argparse.ArgumentParser()
        parser.add_argument("-batch_size", type=int, required=True)
        args = parser.parse_args()

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        run.log_params(args)
        ```
        """
        amplitude.track(amplitude.Event.LOG_PARAMS)

        try:
            # mlfowclient doesnt have log_params api, so we have to use log_batch,
            # This is what internally used by mlflow.log_params
            param_dict = process_params(param_dict)
            param_dict = flatten_dict(param_dict) if flatten_params else param_dict

            params_arr = []
            for param_key in param_dict.keys():
                if len(str(param_key)) > 250 or len(str(param_dict[param_key])) > 250:
                    logger.warning(
                        f"MlFoundry can't log parmeters with length greater than 250 characters. Discarding {param_key}:{param_dict[param_key]}."
                    )
                    continue
                if not self.VALID_PARAM_AND_METRIC_NAMES.match(param_key):
                    logger.warning(
                        f"Invalid param name: {param_key}. Names may only contain alphanumerics, underscores (_), dashes (-), periods (.), spaces ( ), and slashes (/). "
                        f"Discarding param {param_key}={param_dict[param_key]}"
                    )
                    continue
                params_arr.append(Param(param_key, str(param_dict[param_key])))

            if len(params_arr) == 0:
                logger.warning("Cannot log empty params dictionary")

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=params_arr, tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None
        logger.info("Parameters logged successfully")

    def set_tags(self, tags: Dict[str, str]):
        """Set tags for the current `run`.

        Tags are "labels" for a run. A tag is represented by a tag name and value.

        Args:
            tags (Dict[str, str]): A tag name to value map.
                Tag name cannot start with `mlf.`, `mlf.` prefix
                is reserved for mlfoundry. Tag values will be converted
                to `str`.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        run.set_tags({"nlp.framework": "Spark NLP"})

        run.end()
        ```
        """
        amplitude.track(amplitude.Event.SET_TAGS)

        try:
            NAMESPACE.validate_namespace_not_used(names=tags.keys())
            tags_arr = [RunTag(key, str(value)) for key, value in tags.items()]
            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=[], tags=tags_arr
            )
        except MlflowException as e:
            raise MlFoundryException(e.message) from e
        logger.info("Tags set successfully")

    def get_tags(self) -> Dict[str, str]:
        """Returns all the tags set for the current `run`.

        Returns:
            Dict[str, str]: A dictionary containing tags. The keys in the dictionary
                are tag names and the values are corresponding tag values.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        run.set_tags({"nlp.framework": "Spark NLP"})
        print(run.get_tags())

        run.end()
        ```
        """
        amplitude.track(amplitude.Event.GET_TAGS)

        run = self.mlflow_client.get_run(self.run_id)
        return run.data.tags

    def __compute_whylogs_stats(self, df):
        """__compute_whylogs_stats.

        Args:
            df:
        """

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(
                f"df is expected to be a pandas DataFrame but got {str(type(df))}"
            )

        profile_file_name = (
            "profile"
            + "_"
            + datetime.datetime.now().strftime(constants.TIME_FORMAT)
            + ".bin"
        )
        session = whylogs.get_or_create_session()
        profile = session.new_profile()
        profile.track_dataframe(df)
        profile.write_protobuf(profile_file_name)

        try:
            self.mlflow_client.set_tag(self.run_id, "whylogs", True)
            self.mlflow_client.log_artifact(
                self.run_id,
                profile_file_name,
                artifact_path=MlFoundryRun.S3_WHYLOGS_PATH,
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        if os.path.exists(profile_file_name):
            os.remove(profile_file_name)

    def auto_log_metrics(
        self,
        model_type: enums.ModelType,
        data_slice: enums.DataSlice,
        predictions: Collection[Any],
        actuals: Optional[Collection[Any]] = None,
        class_names: Optional[List[str]] = None,
        prediction_probabilities=None,
    ) -> ComputedMetrics:
        """auto_log_metrics.

        Args:
            model_type (enums.ModelType): model_type
            data_slice (enums.DataSlice): data_slice
            predictions (Collection[Any]): predictions
            actuals (Optional[Collection[Any]]): actuals
            class_names (Optional[List[str]]): class_names
            prediction_probabilities:

        Returns:
            ComputedMetrics:
        """
        metrics_calculator = get_metrics_calculator_v2(model_type)
        metrics = metrics_calculator.compute_metrics(
            predictions=predictions,
            actuals=actuals,
            prediction_probabilities=prediction_probabilities,
            class_names=class_names,
        )
        metric_path = os.path.join(constants.ALM_ARTIFACT_PATH, data_slice.value)
        log_artifact_blob(
            mlflow_client=self.mlflow_client,
            run_id=self.run_id,
            blob=metrics.json(),
            file_name=constants.ALM_METRICS_FILE_NAME,
            artifact_path=metric_path,
        )
        return metrics

    def get_metrics(
        self, metric_names: Optional[Iterable[str]] = None
    ) -> Dict[str, List[Metric]]:
        """Get metrics logged for the current `run` grouped by metric name.

        Args:
            metric_names (Optional[Iterable[str]], optional): A list of metric names
                For which the logged metrics will be fetched. If not passed, then all
                metrics logged under the `run` is returned.

        Returns:
            Dict[str, List[Metric]]: A dictionary containing metric name to list of metrics
                map.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        )
        run.log_metrics(metric_dict={"accuracy": 0.7, "loss": 0.6}, step=0)
        run.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4}, step=1)

        metrics = run.get_metrics()
        for metric_name, metric_history in metrics.items():
            print(f"logged metrics for metric {metric_name}:")
            for metric in metric_history:
                print(f"value: {metric.value}")
                print(f"step: {metric.step}")
                print(f"timestamp_ms: {metric.timestamp}")
                print("--")

        run.end()
        ```
        """
        amplitude.track(amplitude.Event.GET_METRICS)
        run = self.mlflow_client.get_run(self.run_id)
        run_metrics = run.data.metrics

        metric_names = (
            set(metric_names) if metric_names is not None else run_metrics.keys()
        )

        unknown_metrics = metric_names - run_metrics.keys()
        if len(unknown_metrics) > 0:
            logger.warning(f"{unknown_metrics} metrics not present in the run")
        metrics_dict = {metric_name: [] for metric_name in unknown_metrics}
        valid_metrics = metric_names - unknown_metrics
        for metric_name in valid_metrics:
            metrics_dict[metric_name] = self.mlflow_client.get_metric_history(
                self.run_id, metric_name
            )
        return metrics_dict

    def get_params(self) -> Dict[str, str]:
        """Get all the params logged for the current `run`.

        Returns:
            Dict[str, str]: A dictionary containing the parameters. The keys in the dictionary
                are parameter names and the values are corresponding parameter values.

        Examples:
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        run.log_params({"learning_rate": 0.01, "epochs": 10})
        print(run.get_params())

        run.end()
        ```
        """
        amplitude.track(amplitude.Event.GET_PARAMS)
        run = self.mlflow_client.get_run(self.run_id)
        return run.data.params

    def log_model(
        self,
        *,
        name: str,
        model: Any,
        framework: Optional[Union[enums.ModelFramework, str]],
        model_save_kwargs: Optional[Dict[str, Any]] = None,
        additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]] = (),
        description: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        model_schema: Optional[Union[Dict[str, Any], ModelSchema]] = None,
        custom_metrics: Optional[List[Union[Dict[str, Any], CustomMetric]]] = None,
        step: int = 0,
    ) -> ModelVersion:
        # TODO (chiragjn): Document mapping of framework to list of valid model save kwargs
        # TODO (chiragjn): Add more examples
        """
        Serialize and log a versioned model under the current project. Each logged model generates a new version
            associated with the given `name` and linked to the current run. Multiple versions of the model can be
            logged as separate versions under the same `name`.

        Args:
            name (str): Name of the model. If a model with this name already exists under the current project,
                the logged model will be added as a new version under that `name`. If no models exist with the given
                `name`, the given model will be logged as version 1.
            model (Any): model instance of any one of the supported frameworks under `mlfoundry.enums.ModelFramework`.
                Can also be `None` which can be useful to create a reference entry without uploading any model files.
            framework (Union[enums.ModelFramework, str]): Model Framework. Ex:- pytorch, sklearn, tensorflow etc.
                The full list of supported frameworks can be found in `mlfoundry.enums.ModelFramework`.
                Can also be `None` when `model` is `None`.
            model_save_kwargs (Optional[Dict[str, Any]], optional): keyword arguments to pass to model serializer.
                Defaults to `None`
            additional_files (Sequence[Tuple[Union[str, Path], Optional[str]]], optional): A list of pairs
                of (source path, destination path) to add additional files and folders
                to the model version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the model versions contents to upload to.
                The model version contents are arranged like follows
                .
                └── model/
                    └── # model files are serialized here
                └── # any additional files and folders can be added here

                To avoid corrupting any model files, the "model/" directory is reserved and additional files cannot be
                added to it.

                E.g. >>> run.log_model(
                     ...     name="xyz", model=clf, framework="sklearn",
                     ...     additional_files=[("foo.txt", "foo/bar/foo.txt"), ("tokenizer/", "foo/tokenizer/")]
                     ... )
                would result in
                .
                ├── model/
                │   └── # model files are serialized here e.g. model.joblib
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            model_schema (Optional[Union[Dict[str, Any], ModelSchema]], optional): instance of `mlfoundry.ModelSchema`.
                This schema needs to be consistent with older versions of the model under the given `name` i.e.
                a feature's value type and model's prediction type cannot be changed in the schema of new version.
                Features can be removed or added between versions.
                E.g. if there exists a v1 with
                schema = {"features": {"name": "feat1": "int"}, "prediction": "categorical"}, then

                schema = {"features": {"name": "feat1": "string"}, "prediction": "categorical"} or
                schema = {"features": {"name": "feat1": "int"}, "prediction": "numerical"}
                are invalid because they change the types of existing features and prediction

                while
                schema = {"features": {"name": "feat1": "int", "feat2": "string"}, "prediction": "categorical"} or
                schema = {"features": {"feat2": "string"}, "prediction": "categorical"}
                are valid

                This field can be updated at any time after logging. Defaults to `None`
            custom_metrics: (Optional[Union[List[Dict[str, Any]], CustomMetric]], optional) = instance of list of `mlfoundry.CustomMetric`
                The custom metrics must be added according to the prediction type of schema.
                custom_metrics = [{
                    "name": "mean_square_error",
                    "type": "metric",
                    "value_type": "float"
                }]
            step (int): step/iteration at which the model is being logged, defaults to 0.

        Returns:
            mlfoundry.ModelVersion: an instance of `ModelVersion` that can be used to download the files,
                load the model, or update attributes like description, metadata, schema.

        Examples:

        ### sklearn
        ```python
        import mlfoundry
        import numpy as np
        from sklearn.pipeline import make_pipeline
        from sklearn.preprocessing import StandardScaler
        from sklearn.svm import SVC

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project"
        )
        X = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
        y = np.array([1, 1, 2, 2])
        clf = make_pipeline(StandardScaler(), SVC(gamma='auto'))
        clf.fit(X, y)

        model_version = run.log_model(
            name="my-sklearn-model",
            model=clf,
            framework="sklearn"
        )
        print(model_version.fqn)
        """
        model_version = _log_model_version(
            run=self,
            name=name,
            model=model,
            framework=framework,
            model_save_kwargs=model_save_kwargs,
            additional_files=additional_files,
            description=description,
            metadata=metadata,
            model_schema=model_schema,
            custom_metrics=custom_metrics,
            step=step,
        )
        logger.info(f"Logged model successfully with fqn {model_version.fqn!r}")
        return model_version

    def log_images(self, images: Dict[str, Image], step: int = 0):
        """Log images under the current `run` at the given `step`.

        Use this function to log images for a `run`. `PIL` package is needed to log images.
        To install the `PIL` package, run `pip install pillow`.

        Args:
            images (Dict[str, "mlfoundry.Image"]): A map of string image key to instance of
                `mlfoundry.Image` class. The image key should only contain alphanumeric,
                hyphens(-) or underscores(_). For a single key and step pair, we can log only
                one image.
            step (int, optional): Training step/iteration for which the `images` should be
                logged. Default is `0`.

        Examples:
        # Logging images from different sources

        ```python
        import mlfoundry
        import numpy as np
        import PIL.Image

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project",
        )

        imarray = np.random.randint(low=0, high=256, size=(100, 100, 3))
        im = PIL.Image.fromarray(imarray.astype("uint8")).convert("RGB")
        im.save("result_image.jpeg")

        images_to_log = {
            "logged-image-array": mlfoundry.Image(data_or_path=imarray),
            "logged-pil-image": mlfoundry.Image(data_or_path=im),
            "logged-image-from-path": mlfoundry.Image(data_or_path="result_image.jpeg"),
        }

        run.log_images(images_to_log, step=1)
        run.end()
        ```
        """
        for key, image in images.items():
            if not isinstance(image, Image):
                raise MlFoundryException("image should be of type `mlfoundry.Image`")
            image.save(run=self, key=key, step=step)

    def log_plots(
        self,
        plots: Dict[
            str,
            Union[
                "matplotlib.pyplot",
                "matplotlib.figure.Figure",
                "plotly.graph_objects.Figure",
                Plot,
            ],
        ],
        step: int = 0,
    ):
        """Log custom plots under the current `run` at the given `step`.

        Use this function to log custom matplotlib, plotly plots.

        Args:
            plots (Dict[str, "matplotlib.pyplot", "matplotlib.figure.Figure", "plotly.graph_objects.Figure", Plot]):
                A map of string plot key to the plot or figure object.
                The plot key should only contain alphanumeric, hyphens(-) or
                underscores(_). For a single key and step pair, we can log only
                one image.
            step (int, optional): Training step/iteration for which the `plots` should be
                logged. Default is `0`.


        Examples:
        ### Logging a plotly figure
        ```python
        import mlfoundry
        import plotly.express as px

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project",
        )

        df = px.data.tips()
        fig = px.histogram(
            df,
            x="total_bill",
            y="tip",
            color="sex",
            marginal="rug",
            hover_data=df.columns,
        )

        plots_to_log = {
            "distribution-plot": fig,
        }

        run.log_plots(plots_to_log, step=1)
        run.end()
        ```

        ### Logging a matplotlib plt or figure
        ```python
        import mlfoundry
        from matplotlib import pyplot as plt
        import numpy as np

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project",
        )

        t = np.arange(0.0, 5.0, 0.01)
        s = np.cos(2 * np.pi * t)
        (line,) = plt.plot(t, s, lw=2)

        plt.annotate(
            "local max",
            xy=(2, 1),
            xytext=(3, 1.5),
            arrowprops=dict(facecolor="black", shrink=0.05),
        )

        plt.ylim(-2, 2)

        plots_to_log = {"cos-plot": plt, "cos-plot-using-figure": plt.gcf()}

        run.log_plots(plots_to_log, step=1)
        run.end()
        ```
        """
        for key, plot in plots.items():
            plot = Plot(plot) if not isinstance(plot, Plot) else plot
            plot.save(run=self, key=key, step=step)

    def use_model(self, fqn: str) -> ModelVersion:
        """
        Get the model version to download contents or load it in memory

        Args:
            fqn (str): Fully qualified name of the model version.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(project_name="my-new-project")
            model_version = run.use_model(fqn="model:truefoundry/user/my-classification-project/my-sklearn-model:1")

            # load the model into memory
            clf = model_version.load()

            # download the model to disk
            temp = tempfile.TemporaryDirectory()
            download_info = model_version.download(path=temp.name)
            print(download_info)
            ```
        """
        # TODO (chiragjn): implement event logging
        return ModelVersion(fqn)

    def use_artifact(self, fqn: str) -> ArtifactVersion:
        """
        Get the artifact version to download contents

        Args:
            fqn (str): Fully qualified name of the artifact version.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            run = client.create_run(project_name="my-new-project")
            artifact_version = run.use_artifact(fqn="artifact:truefoundry/user/my-classification-project/sklearn-artifact:1")

            # download the artifact to disk
            temp = tempfile.TemporaryDirectory()
            download_info = artifact_version.download(path=temp.name)
            print(download_info)
            ```
        """
        # TODO (chiragjn): implement event logging
        return ArtifactVersion(fqn)



from __future__ import annotations

import threading
from functools import lru_cache, wraps
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import click
from filelock import FileLock, Timeout
from pydantic import BaseModel, Field, constr

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger
from mlfoundry.run_utils import resolve_tracking_uri
from mlfoundry.tracking.entities import Token
from mlfoundry.tracking.truefoundry_rest_store import get_rest_store

CREDENTIALS_DIR = Path.home() / ".truefoundry"
CREDENTIALS_FILE = CREDENTIALS_DIR / "credentials.json"


OLD_CREDENTIALS_DIR = Path.home() / ".mlfoundry"
OLD_CREDENTIALS_FILE = OLD_CREDENTIALS_DIR / "credentials.netrc"

if OLD_CREDENTIALS_FILE.exists():
    logger.warning(
        "%s file is deprecated. You can delete this file now.", OLD_CREDENTIALS_FILE
    )


class CredentialsFileContent(BaseModel):
    access_token: constr(min_length=1) = Field(repr=False)
    refresh_token: constr(min_length=1) = Field(repr=False)
    host: constr(min_length=1)

    class Config:
        allow_mutation = False

    def to_token(self) -> Token:
        return Token(access_token=self.access_token, refresh_token=self.refresh_token)


def _ensure_lock_taken(method):
    @wraps(method)
    def lock_guard(self, *method_args, **method_kwargs):
        if not self.lock_taken():
            raise Exception(
                "Trying to write to credential file without using with block"
            )
        return method(self, *method_args, **method_kwargs)

    return lock_guard


CRED_FILE_THREAD_LOCK = threading.RLock()


@lru_cache(maxsize=None)
def get_file_lock(lock_file_path: str) -> FileLock:
    return FileLock(lock_file_path)


class CredentialsFileManager:
    def __init__(
        self,
        credentials_file_path: Path = CREDENTIALS_FILE,
        lock_timeout: float = 60.0,
    ):
        credentials_file_path = credentials_file_path.absolute()

        logger.debug("credential file path %r", credentials_file_path)

        credentials_lock_file_path = f"{credentials_file_path}.lock"

        logger.debug("credential lock file path %r", credentials_lock_file_path)
        self._credentials_file_path = credentials_file_path

        cred_file_dir = credentials_file_path.parent
        cred_file_dir.mkdir(exist_ok=True, parents=True)

        self._file_lock = get_file_lock(credentials_lock_file_path)
        self._lock_timeout = lock_timeout
        self._lock_owner: Optional[int] = None

    def __enter__(self) -> CredentialsFileManager:
        # The lock objects are recursive locks, which means that once acquired, they will not block on successive lock requests:
        lock_aquired = CRED_FILE_THREAD_LOCK.acquire(timeout=self._lock_timeout)
        if not lock_aquired:
            raise MlFoundryException(
                "Could not aquire CRED_FILE_THREAD_LOCK"
                f" in {self._lock_timeout} seconds"
            )
        try:
            self._file_lock.acquire(timeout=self._lock_timeout)
        except Timeout as ex:
            raise MlFoundryException(
                f"Failed to aquire lock on credential file within {self._lock_timeout} seconds.\n"
                "Is any other process trying to login?"
            ) from ex
        logger.debug("Acquired file and thread lock to access credential file")
        self._lock_owner = threading.get_ident()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._file_lock.release()
        CRED_FILE_THREAD_LOCK.release()
        logger.debug("Released file and thread lock to access credential file")
        self._lock_owner = None

    def lock_taken(self) -> bool:
        return self._lock_owner == threading.get_ident()

    @_ensure_lock_taken
    def read(self) -> CredentialsFileContent:
        try:
            return CredentialsFileContent.parse_file(self._credentials_file_path)
        except Exception as ex:
            raise MlFoundryException(
                "Error while reading the credentials file "
                f"{self._credentials_file_path}. Please login again "
                "using `mlfoundry login` command or `mlfoundry.login()` function"
            ) from ex

    @_ensure_lock_taken
    def write(self, credentials_file_content: CredentialsFileContent):
        if not isinstance(credentials_file_content, CredentialsFileContent):
            raise MlFoundryException(
                "Only object of type `CredentialsFileContent` is allowed. "
                f"Got {type(credentials_file_content)}"
            )
        logger.debug("Updating the credential file content")
        with open(self._credentials_file_path, "w", encoding="utf8") as file:
            file.write(credentials_file_content.json())

    @_ensure_lock_taken
    def exists(self) -> bool:
        return self._credentials_file_path.exists()


def login(
    tracking_uri: Optional[str] = None,
    relogin: bool = False,
    api_key: Optional[str] = None,
) -> bool:
    """Save API key in local file system for a given `tracking_uri`.

    Args:
        tracking_uri (Optional[str], optional): tracking_uri for the given API key
        relogin (bool, optional): Overwrites the existing API key for the `tracking_uri` if
            set to `True`. If set to `False` and an API key is already present for
            the given `tracking_uri`, then the existing API key is kept untouched.
            Default is `False`.
        api_key (Optional[str], optional): The API key for the given `tracking_uri`.
            If `api_key` is not passed, this function prompts for the API key.

    Returns:
        bool: Returns `True` if any credential was persisted.
    """
    from mlfoundry.session import EnvCredentialProvider

    if EnvCredentialProvider.can_provide():
        logger.warning(
            "TFY_API_KEY env var is already set. "
            "When running mlfoundry, it will use the api key to authorize.\n"
            "Login will just save the credentials on disk."
        )

    tracking_uri = resolve_tracking_uri(tracking_uri).strip("/")
    auth_service = get_rest_store(tracking_uri).get_auth_service()

    cred_file = CredentialsFileManager()

    with cred_file:
        if not relogin and cred_file.exists():
            cred_file_content = cred_file.read()
            if tracking_uri != cred_file_content.host:
                if click.confirm(
                    f"Already logged in to {cred_file_content.host!r}\n"
                    f"Do you want to relogin to {tracking_uri!r}?"
                ):
                    return login(
                        tracking_uri=tracking_uri, relogin=True, api_key=api_key
                    )
            user_info = cred_file_content.to_token().to_user_info()
            print(
                f"Already logged in to {cred_file_content.host!r} as "
                f"{user_info.user_id!r} ({user_info.email!r})\n"
                "Please use `mlfoundry login --relogin` or `mlfoundry.login(relogin=True)` "
                "to force relogin"
            )
            return False

        if api_key:
            logger.debug("Logging in with api key")
            token = auth_service.get_token_from_api_key(api_key=api_key)
            cred_file_content = CredentialsFileContent(
                access_token=token.access_token,
                refresh_token=token.refresh_token,
                host=tracking_uri,
            )
            cred_file.write(cred_file_content)
        else:
            device_code = auth_service.get_device_code()
            url_to_go = device_code.get_user_clickable_url(tracking_uri=tracking_uri)
            print(f"Opening:- {url_to_go}")
            print(
                "Please click on the above link if it is not "
                "automatically opened in a browser window."
            )
            click.launch(url_to_go)
            token = auth_service.get_token_from_device_code(
                device_code=device_code.device_code, timeout=120
            )
            cred_file_content = CredentialsFileContent(
                access_token=token.access_token,
                refresh_token=token.refresh_token,
                host=tracking_uri,
            )
            cred_file.write(cred_file_content)

    user_info = token.to_user_info()
    print(
        f"logged in to {cred_file_content.host!r} as {user_info.user_id!r} ({user_info.email})"
    )
    return True



import os

import click

from mlfoundry.cli.commands import download, login
from mlfoundry.logger import logger
from mlfoundry.version import __version__


@click.group()
@click.version_option(__version__)
def mlfoundry_cli():
    """MlFoundry CLI"""
    click.secho("MlFoundry CLI", bold=True, fg="green")


def create_mlfoundry_cli():
    """Generates CLI by combining all subcommands into a main CLI and returns in
    Returns:
        function: main CLI functions will all added sub-commands
    """
    _cli = mlfoundry_cli
    _cli.add_command(ui)
    _cli.add_command(download)
    _cli.add_command(login)
    return _cli()


@click.command(
    help="Generate MLFoundry Dashboard",
    short_help="Generate MLFoundry Dashboard",
)
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True, dir_okay=True, readable=True),
    default=os.path.abspath("."),
)
def ui(path: str):
    os.system(f"mlfoundry_ui start-dashboard -p {path}")



import re

IMAGE_METADATA_FILE_NAME = "image_metadata.json"
DEFAULT_IMAGE_FORMAT = "png"
MISSING_PILLOW_PACKAGE_MESSAGE = (
    "We need PIL package to save image.\nTo install, run `pip install pillow`"
)
IMAGE_KEY_REGEX = re.compile(r"^[a-zA-Z0-9-_]+$")



import enum
import json
import os
import posixpath
import sys
import tempfile
from typing import Union
from uuid import uuid4

from mlflow.entities import ArtifactType
from pydantic import BaseModel

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.artifact import (
    ArtifactVersionInternalMetadata,
    _log_artifact_version_helper,
)
from mlfoundry.log_types.artifacts.constants import FILES_DIR, INTERNAL_METADATA_PATH
from mlfoundry.log_types.pydantic_base import PydanticBase
from mlfoundry.log_types.utils import validate_key_name

PlotObjType = Union[
    "matplotlib.figure.Figure",
    "plotly.graph_objects.Figure",
    "matplotlib.pyplot",
]


@enum.unique
class Format(enum.Enum):
    SVG = "SVG"
    HTML = "HTML"
    PNG = "PNG"


def _is_matplotlib_figure(fig) -> bool:
    if "matplotlib" not in sys.modules:
        return False
    import matplotlib

    return isinstance(fig, matplotlib.figure.Figure)


def _is_matplotlib_plt(plt) -> bool:
    if "matplotlib" not in sys.modules:
        return False
    return getattr(plt, "__name__", "") == "matplotlib.pyplot"


def _is_plotly_figure(fig) -> bool:
    if "plotly" not in sys.modules:
        return False

    import plotly

    return isinstance(fig, plotly.graph_objects.Figure)


def get_plot_file_name(format: Format) -> str:
    return f"plot.{format.value.lower()}"


class PlotArtifact(BaseModel):
    artifact_file: str
    format: Format

    class Config:
        allow_mutation = False
        use_enum_values = True


def _save_matplotlib_figure(
    figure: "matplotlib.figure.Figure",
    key: str,
    step: int,
    local_dir: str,
) -> PlotArtifact:
    supported_formats = figure.canvas.get_supported_filetypes().keys()
    if "svg" in supported_formats:
        format_ = Format.SVG
    elif "png" in supported_formats:
        format_ = Format.PNG
    else:
        raise MlFoundryException(
            f"Could not save {key} {figure} matplotlib figure"
            "in either SVG or PNG format"
        )
    file_path = get_plot_file_name(format=format_)
    local_path = os.path.join(local_dir, file_path)
    figure.savefig(local_path)
    return PlotArtifact(artifact_file=file_path, format=format_)


def _save_matplotlib_plt(
    plt: "matplotlib.pyplot",
    key: str,
    step: int,
    local_dir: str,
) -> PlotArtifact:
    figure = plt.gcf()
    return _save_matplotlib_figure(
        figure=figure, key=key, step=step, local_dir=local_dir
    )


def _save_plotly_figure(
    figure: "plotly.graph_objects.Figure",
    key: str,
    step: int,
    local_dir: str,
) -> PlotArtifact:
    format_ = Format.HTML
    file_path = get_plot_file_name(format=format_)
    local_path = os.path.join(local_dir, file_path)
    figure.write_html(local_path, include_plotlyjs="cdn", auto_open=False)
    return PlotArtifact(artifact_file=file_path, format=format_)


class Plot:
    def __init__(self, plot_obj: PlotObjType):
        self._plot_obj = plot_obj

    def _save_plot(self, key: str, step: int, local_dir: str) -> PlotArtifact:
        if _is_matplotlib_plt(self._plot_obj):
            return _save_matplotlib_plt(
                plt=self._plot_obj, key=key, step=step, local_dir=local_dir
            )

        if _is_matplotlib_figure(self._plot_obj):
            return _save_matplotlib_figure(
                figure=self._plot_obj, key=key, step=step, local_dir=local_dir
            )

        if _is_plotly_figure(self._plot_obj):
            return _save_plotly_figure(
                figure=self._plot_obj, key=key, step=step, local_dir=local_dir
            )

        raise MlFoundryException(
            f"Unknown type: {type(self._plot_obj)}"
            "Supported types are, matplotlib.figure.Figure, matplotlib.pyplot"
            " and plotly.graph_objects.Figure"
        )

    def save(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
    ):
        validate_key_name(key)

        # creating a temp dir which will be logged
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")
        local_files_dir = os.path.join(temp_dir.name, FILES_DIR)
        os.makedirs(local_files_dir, exist_ok=True)

        # save plot locally
        plot_artifact = self._save_plot(key=key, step=step, local_dir=local_files_dir)

        # save internal metadata
        internal_metadata = PlotVersionInternalMetadata(
            files_dir=FILES_DIR,
            plot_file=posixpath.join(FILES_DIR, plot_artifact.artifact_file),
        )
        local_internal_metadata_path = os.path.join(
            temp_dir.name, INTERNAL_METADATA_PATH
        )
        os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)
        with open(local_internal_metadata_path, "w") as f:
            json.dump(internal_metadata.dict(), f)

        # log the artifact
        _log_artifact_version_helper(
            run=run,
            name=key,
            artifact_type=ArtifactType.PLOT,
            artifact_dir=temp_dir,
            step=step,
        )


class PlotRunLogType(PydanticBase):
    value: PlotArtifact

    @staticmethod
    def get_log_type() -> str:
        return "plot"


class PlotVersionInternalMetadata(ArtifactVersionInternalMetadata):
    plot_file: str


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    import plotly.express as px

    import mlfoundry

    client = mlfoundry.get_client("https://app.devtest.truefoundry.tech")
    # client = mlfoundry.get_client("http://localhost:5000")

    run = client.create_run(project_name="plot-test1")

    df = px.data.iris()
    fig = px.scatter(
        df,
        x="sepal_width",
        y="sepal_length",
        color="species",
        size="petal_length",
        hover_data=["petal_width"],
    )
    Plot(fig).save(run, "foo")

    df = px.data.tips()
    fig = px.histogram(
        df,
        x="total_bill",
        y="tip",
        color="sex",
        marginal="rug",
        hover_data=df.columns,
    )
    Plot(fig).save(run, "foo", step=1)

    names = ["group_a", "group_b", "group_c"]
    values = [1, 10, 100]
    plt.figure(figsize=(9, 3))
    plt.subplot(131)
    plt.bar(names, values)
    plt.subplot(132)
    plt.scatter(names, values)
    plt.subplot(133)
    plt.plot(names, values)
    plt.suptitle("Categorical Plotting")

    Plot(plt).save(run, "bar")

    plt.clf()

    data = {
        "a": np.arange(50),
        "c": np.random.randint(0, 50, 50),
        "d": np.random.randn(50),
    }
    data["b"] = data["a"] + 10 * np.random.randn(50)
    data["d"] = np.abs(data["d"]) * 100
    plt.scatter("a", "b", c="c", s="d", data=data)
    plt.xlabel("entry a")
    plt.ylabel("entry b")
    Plot(plt).save(run, "bar", 2)

    plt.clf()

    ax = plt.subplot()
    t = np.arange(0.0, 5.0, 0.01)
    s = np.cos(2 * np.pi * t)
    (line,) = plt.plot(t, s, lw=2)

    plt.annotate(
        "local max",
        xy=(2, 1),
        xytext=(3, 1.5),
        arrowprops=dict(facecolor="black", shrink=0.05),
    )

    plt.ylim(-2, 2)
    Plot(plt.gcf()).save(run, "bar", 3)



import copy
import json
import os
import tempfile
from pathlib import Path
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union

from mlflow.entities import Artifact, ArtifactType
from mlflow.tracking import MlflowClient
from pydantic import BaseModel, Extra

from mlfoundry.artifact.truefoundry_artifact_repo import MlFoundryArtifactsRepository
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.constants import INTERNAL_METADATA_PATH
from mlfoundry.log_types.artifacts.utils import (
    _get_mlflow_client,
    _validate_artifact_metadata,
    _validate_description,
)
from mlfoundry.logger import logger

# TODO: Add some progress indicators for upload and download
# TODO: Support async download and upload


class ArtifactPath(NamedTuple):
    src: str
    dest: str = None


class ArtifactVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root


class ArtifactVersionDownloadInfo(BaseModel):
    download_dir: str
    content_dir: str


class ArtifactVersion:
    def __init__(self, fqn: str) -> None:
        self._mlflow_client = _get_mlflow_client()
        self._artifact_version = self._mlflow_client.get_artifact_version_by_fqn(
            fqn=fqn
        )
        self._artifact: Artifact = self._mlflow_client.get_artifact_by_id(
            artifact_id=self._artifact_version.artifact_id
        )
        self._mlfa_repo = MlFoundryArtifactsRepository(
            version_id=self._artifact_version.id, mlflow_client=self._mlflow_client
        )
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._set_mutable_attrs()

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Artifact Version was deleted, cannot access a deleted version"
            )

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            self._artifact_version = self._mlflow_client.get_artifact_version_by_id(
                version_id=self._artifact_version.id
            )
        self._description = self._artifact_version.description or ""
        self._metadata = copy.deepcopy(self._artifact_version.artifact_metadata)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    @property
    def name(self) -> str:
        return self._artifact.name

    @property
    def artifact_fqn(self) -> str:
        return self._artifact.fqn

    @property
    def version(self) -> int:
        return self._artifact_version.version

    @property
    def fqn(self) -> str:
        return self._artifact_version.fqn

    @property
    def step(self) -> Optional[int]:
        return self._artifact_version.step

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, value: str):
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        _validate_artifact_metadata(value)
        self._metadata = value

    def raw_download(self, path: Optional[Union[str, Path]]) -> str:
        logger.info(
            "Downloading artifact version contents, this might take a while ..."
        )
        return self._mlfa_repo.download_artifacts(artifact_path="", dst_path=path)

    def _download(
        self, path: Optional[Union[str, Path]]
    ) -> Tuple[ArtifactVersionInternalMetadata, str]:
        self._ensure_not_deleted()
        download_dir = self.raw_download(path=path)
        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                f"Artifact version seems to be corrupted or in invalid format due to missing artifact metadata. "
                f"You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ArtifactVersionInternalMetadata.parse_obj(json.load(f))
        download_path = os.path.join(download_dir, internal_metadata.files_dir)
        return internal_metadata, download_path

    def download(self, path: Optional[Union[str, Path]]) -> str:
        _, download_path = self._download(path=path)
        return download_path

    def delete(self) -> bool:
        self._ensure_not_deleted()
        self._mlflow_client.delete_artifact_version(
            version_id=self._artifact_version.id
        )
        self._deleted = True
        return True

    def update(self):
        self._ensure_not_deleted()

        self._artifact_version = self._mlflow_client.update_artifact_version(
            version_id=self._artifact_version.id,
            description=self.description,
            artifact_metadata=self.metadata,
        )
        self._set_mutable_attrs()


def _log_artifact_version_helper(
    run,
    name: str,
    artifact_type: ArtifactType,
    artifact_dir: tempfile.TemporaryDirectory,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: int = 0,
):
    mlflow_client: MlflowClient = run.mlflow_client
    version_id = mlflow_client.create_artifact_version(
        experiment_id=int(run._experiment_id),
        artifact_type=artifact_type,
        name=name,
    )
    artifacts_repo = MlFoundryArtifactsRepository(
        version_id=version_id, mlflow_client=mlflow_client
    )
    try:
        logger.info("Packaging and uploading files to remote ...")
        artifacts_repo.log_artifacts(local_dir=artifact_dir.name, artifact_path=None)
    except Exception as e:
        mlflow_client.notify_failure_for_artifact_version(version_id=version_id)
        raise MlFoundryException("Failed to log Artifact") from e
    finally:
        artifact_dir.cleanup()
    artifact_version = mlflow_client.finalize_artifact_version(
        version_id=version_id,
        run_uuid=run.run_id,
        description=description,
        artifact_metadata=metadata,
        data_path=INTERNAL_METADATA_PATH,
        step=step,
    )

    return ArtifactVersion(fqn=artifact_version.fqn)



"""
.
├── .truefoundry/
│   └── metadata.json
└── files/
    └── model/
"""
import posixpath

INTERNAL_METADATA_DIR = ".truefoundry"
INTERNAL_METADATA_FILE_NAME = "metadata.json"
INTERNAL_METADATA_PATH = posixpath.join(
    INTERNAL_METADATA_DIR, INTERNAL_METADATA_FILE_NAME
)
FILES_DIR = "files"
MODEL_DIR_NAME = "model"
DESCRIPTION_MAX_LENGTH = 1024

# Link to docs here explaining schema consistency across model versions
MODEL_SCHEMA_UPDATE_FAILURE_HELP = """Model was logged successfully but failed to update the model schema because
it is inconsistent with the previous versions of the model. You can still fix the schema and update it using the
following:

```
import mlfoundry
from mlfoundry import ModelVersion, ModelSchema, Feature, FeatureValueType, PredictionValueType
client = mlfoundry.get_client()
model_version = ModelVersion(fqn="{fqn}")
model_version = ModelSchema(...) # or schema in dictionary format {{"features": [...], "prediction": ...}}
model_version.update()
```
"""



import abc
import typing

import numpy as np
from pydantic import BaseModel

from mlfoundry.enums import ModelType
from mlfoundry.metrics.v2.utils import get_class_names


class ComputedMetrics(BaseModel):
    model_type: str
    metrics: typing.Dict[str, typing.Any]


class BaseMetrics(abc.ABC):
    @property
    @abc.abstractmethod
    def model_type(self) -> ModelType:
        pass

    @abc.abstractmethod
    def get_metric_name_to_function_map(self) -> typing.Dict[str, typing.Callable]:
        pass

    def validate(
        self,
        predictions,
        actuals,
        prediction_probabilities,
        class_names,
    ):
        pass

    def compute_metrics(
        self,
        predictions=None,
        actuals=None,
        prediction_probabilities=None,
        class_names=None,
    ) -> ComputedMetrics:
        self.validate(
            predictions=predictions,
            actuals=actuals,
            prediction_probabilities=prediction_probabilities,
            class_names=class_names,
        )
        metric_name_to_function_map = self.get_metric_name_to_function_map()

        if class_names is None:
            class_names = get_class_names(predictions=predictions, actuals=actuals)
        elif isinstance(class_names, np.ndarray):
            class_names = class_names.tolist()

        metrics = {}
        for (
            metric_name,
            metric_function,
        ) in metric_name_to_function_map.items():
            metric_value = metric_function(
                predictions=predictions,
                actuals=actuals,
                prediction_probabilities=prediction_probabilities,
                class_names=class_names,
            )
            if metric_value is None:
                continue
            metrics[metric_name] = metric_value

        return ComputedMetrics(model_type=self.model_type.value, metrics=metrics)



import typing

import pandas as pd
from mlflow.tracking import MlflowClient

from mlfoundry.dataset.schema import build_schema
from mlfoundry.dataset.serde import TabularDatasetSerDe
from mlfoundry.dataset.stats import build_stats
from mlfoundry.dataset.types import DataSet, DatasetSchema, DatasetStats, Profiles
from mlfoundry.dataset.validation import validate_dataset, validate_dataset_name
from mlfoundry.exceptions import MlFoundryException


def pre_process_features(features) -> pd.DataFrame:
    try:
        features_df = pd.DataFrame(features).reset_index(drop=True)
        features_df.columns = features_df.columns.map(str)
        return features_df
    except Exception as ex:
        raise MlFoundryException("could not convert features to a DataFrame") from ex


def pre_process_actuals(actuals) -> pd.Series:
    try:
        return pd.Series(actuals).reset_index(drop=True)
    except Exception as ex:
        raise MlFoundryException("could not convert actuals to a Series") from ex


def pre_process_predictions(predictions) -> pd.Series:
    try:
        return pd.Series(predictions).reset_index(drop=True)
    except Exception as ex:
        raise MlFoundryException("could not convert predictions to a Series") from ex


class TabularDatasetDriver:
    """TabularDataset."""

    def __init__(self, mlflow_client: MlflowClient, run_id: str):
        self.mlflow_client: MlflowClient = mlflow_client
        self._serde: TabularDatasetSerDe = TabularDatasetSerDe(self.mlflow_client)
        self._run_id: str = run_id

    def log_dataset(
        self,
        dataset_name: str,
        features,
        predictions=None,
        actuals=None,
        only_stats: bool = False,
    ):
        """
        Log a dataset associated with a run. A dataset is a collection of features,
        predictions and actuals. Datasets are uniquely identified by the dataset_name
        under a run. They are immutable, once successfully logged, overwriting it is not allowed.

        Mixed types are not allowed in features, actuals and predictions. However, there can be
        missing data in the form of None, NaN, NA.

        :param dataset_name:    Name of the dataset. Dataset name should only contain letters,
                                numbers, underscores and hyphens.
        :type dataset_name: str
        :param features:        Features associated with this dataset.
                                This should be either pandas DataFrame or should be of a
                                data type which can be convered to a DataFrame.
        :param predictions:     Predictions associated with this dataset and run. This
                                should be either pandas Series or should be of a data type which
                                can be convered to a DataFrame. This is an optional argument.
        :param actuals:         Actuals associated with this dataset and run. This
                                should be either pandas Series or should be of a data type which
                                can be convered to a DataFrame. This is an optional argument.
        :param only_stats:      If True, then the dataset (features, predictions, actuals) is
                                not saved. Only statistics and the dataset schema will be
                                persisted. Default is False.
        :type only_stats: bool
        """
        validate_dataset_name(dataset_name)
        self._serde.validate_dataset_is_not_already_logged(
            run_id=self._run_id, dataset_name=dataset_name
        )

        features = pre_process_features(features)
        predictions = (
            predictions if predictions is None else pre_process_predictions(predictions)
        )
        actuals = actuals if actuals is None else pre_process_actuals(actuals)

        schema = build_schema(
            features=features, predictions=predictions, actuals=actuals
        )
        validate_dataset(
            features=features, predictions=predictions, actuals=actuals, schema=schema
        )
        dataset_stats, profiles = build_stats(
            features=features, predictions=predictions, actuals=actuals
        )
        self._serde.save_dataset(
            run_id=self._run_id,
            dataset_name=dataset_name,
            features=features,
            predictions=predictions,
            actuals=actuals,
            schema=schema,
            stats=dataset_stats,
            profiles=profiles,
            only_stats=only_stats,
        )

    def get_dataset(self, dataset_name: str) -> typing.Optional[DataSet]:
        """
        Returns the features, predictions, actuals associated with the dataset.
        If only_stats was set to True while logging the dataset,
        this will return None.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: typing.Optional[DataSet]
        """
        features = self.get_features(dataset_name)
        if features is None:
            return None
        return DataSet(
            dataset_name=dataset_name,
            features=features,
            actuals=self.get_actuals(dataset_name),
            predictions=self.get_predictions(dataset_name),
        )

    def get_features(self, dataset_name: str) -> typing.Optional[pd.DataFrame]:
        """
        Returns the features associated with the dataset.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: typing.Optional[pd.DataFrame]
        """
        return self._serde.load_features(run_id=self._run_id, dataset_name=dataset_name)

    def get_predictions(self, dataset_name: str) -> typing.Optional[pd.Series]:
        """
        Returns the predictions associated with the dataset.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: typing.Optional[pd.Series]
        """
        return self._serde.load_predictions(
            run_id=self._run_id, dataset_name=dataset_name
        )

    def get_actuals(self, dataset_name: str) -> typing.Optional[pd.Series]:
        """
        Returns the actuals associated with the dataset.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: typing.Optional[pd.Series]
        """
        return self._serde.load_actuals(run_id=self._run_id, dataset_name=dataset_name)

    def get_schema(self, dataset_name: str) -> DatasetSchema:
        """
        Returns the schema generated from the dataset.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: DatasetSchema
        """
        return self._serde.load_schema(run_id=self._run_id, dataset_name=dataset_name)

    def get_stats(self, dataset_name: str) -> DatasetStats:
        """
        Returns the statistics generated from the dataset.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: DatasetStats
        """
        return self._serde.load_stats(run_id=self._run_id, dataset_name=dataset_name)

    def get_profiles(self, dataset_name: str) -> Profiles:
        """
        Returns the Whylogs DatasetProfile generated from the dataset.
        There will be seperate DatasetProfiles for features, predictions and actuals

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: Profiles
        """
        return self._serde.load_profiles(run_id=self._run_id, dataset_name=dataset_name)



import collections
import json
import os.path
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union

from mlflow.entities import ArtifactType

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.artifact import (
    ArtifactPath,
    ArtifactVersion,
    ArtifactVersionInternalMetadata,
    _log_artifact_version_helper,
)
from mlfoundry.log_types.artifacts.constants import FILES_DIR, INTERNAL_METADATA_PATH
from mlfoundry.log_types.artifacts.utils import (
    _copy_additional_files,
    _validate_artifact_metadata,
    _validate_description,
)
from mlfoundry.logger import logger


def _log_artifact_version(
    run,
    name: str,
    artifact_paths: List[Union[ArtifactPath, Tuple[str, Optional[str]]]],
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: int = 0,
) -> ArtifactVersion:

    for i, artifact_path in enumerate(artifact_paths):
        if isinstance(artifact_path, ArtifactPath):
            continue
        elif isinstance(artifact_path, collections.Sequence) and (
            0 < len(artifact_path) <= 2
        ):
            artifact_paths[i] = ArtifactPath(*artifact_path)
        else:
            raise ValueError(
                "`artifact_path` should be an instance of `mlfoundry.ArtifactPath` or a tuple of (src, dest) path strings"
            )

    metadata = metadata or {}
    step = step or 0

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    logger.info("Logging the artifact, this might take a while ...")
    temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

    internal_metadata = ArtifactVersionInternalMetadata(
        files_dir=FILES_DIR,
    )

    try:
        local_files_dir = os.path.join(temp_dir.name, internal_metadata.files_dir)
        os.makedirs(local_files_dir, exist_ok=True)

        logger.info("Copying the files to log")
        _copy_additional_files(
            root_dir=temp_dir.name,
            files_dir=internal_metadata.files_dir,
            model_dir=None,
            additional_files=artifact_paths,
        )

    # TODO(nikp1172) verify error messag when artifact doesn't exist
    except Exception as e:
        temp_dir.cleanup()
        raise MlFoundryException("Failed to log artifact") from e

    # save internal metadata
    local_internal_metadata_path = os.path.join(temp_dir.name, INTERNAL_METADATA_PATH)
    os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)
    with open(local_internal_metadata_path, "w") as f:
        json.dump(internal_metadata.dict(), f)

    return _log_artifact_version_helper(
        run=run,
        name=name,
        artifact_type=ArtifactType.ARTIFACT,
        artifact_dir=temp_dir,
        description=description,
        metadata=metadata,
        step=step,
    )



import time
from typing import Dict, Optional
from urllib.parse import urlparse

import jwt
from pydantic import BaseModel, Field, constr, validator


class AuthServerInfo(BaseModel):
    tenant_name: constr(min_length=1)
    auth_server_url: str

    class Config:
        allow_mutation = False


class ArtifactCredential(BaseModel):
    run_id: str
    path: str
    signed_uri: str


class UserInfo(BaseModel):
    user_id: constr(min_length=1)
    email: str
    tenant_name: constr(min_length=1) = Field(alias="tenantName")

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False


class Token(BaseModel):
    access_token: constr(min_length=1) = Field(alias="accessToken", repr=False)
    refresh_token: constr(min_length=1) = Field(alias="refreshToken", repr=False)
    decoded_value: Optional[Dict] = Field(exclude=True, repr=False)

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    @validator("decoded_value", always=True, pre=True)
    def _decode_jwt(cls, v, values, **kwargs):
        access_token = values["access_token"]
        return jwt.decode(
            access_token,
            options={
                "verify_signature": False,
                "verify_aud": False,
                "verify_exp": False,
            },
        )

    @property
    def tenant_name(self) -> str:
        return self.decoded_value["tenantName"]

    def is_going_to_be_expired(self, buffer_in_seconds: int = 120) -> bool:
        exp = int(self.decoded_value["exp"])
        return (exp - time.time()) < buffer_in_seconds

    def to_user_info(self) -> UserInfo:
        return UserInfo(
            user_id=self.decoded_value["username"],
            email=self.decoded_value["email"],
            tenant_name=self.tenant_name,
        )


class DeviceCode(BaseModel):
    user_code: str = Field(alias="userCode")
    device_code: str = Field(alias="deviceCode")

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    def get_user_clickable_url(self, tracking_uri: str) -> str:
        parsed_tracking_uri = urlparse(tracking_uri)
        return f"{parsed_tracking_uri.scheme}://{parsed_tracking_uri.netloc}/authorize/device?userCode={self.user_code}"



"""
# TO independently test this module, you can run the example in the path
python examples/sklearn/iris_train.py

Besides running pytest
"""
import os
import urllib
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Generator, List, Optional, Sequence, Union

import coolname
import mlflow
import pandas as pd
from mlflow.store.tracking import SEARCH_MAX_RESULTS_DEFAULT
from mlflow.tracking import MlflowClient

from mlfoundry import amplitude, constants, env_vars
from mlfoundry.enums import ViewType
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.log_types.artifacts.artifact import ArtifactVersion
from mlfoundry.log_types.artifacts.model import ModelVersion
from mlfoundry.logger import logger
from mlfoundry.mlfoundry_run import MlFoundryRun
from mlfoundry.monitoring.entities import Actual, Prediction
from mlfoundry.monitoring.store import MonitoringClient
from mlfoundry.session import Session, init_session


def _get_internal_env_vars_values() -> Dict[str, str]:
    env = {}
    for env_var_name in env_vars.INTERNAL_ENV_VARS:
        value = os.getenv(env_var_name)
        if value:
            env[env_var_name] = value

    return env


def get_client(
    tracking_uri: Optional[str] = None,
    disable_analytics: bool = False,
    api_key: Optional[str] = None,
) -> "MlFoundry":
    """Initializes and returns the mlfoundry client.

    Args:
        tracking_uri(deprecated) (Optional[str], optional): Custom tracking server URL.
            If not passed, by default all the run details are sent to Truefoundry server
            and can be visualized at https://app.truefoundry.com/mlfoundry.
            Tracking server URL can be also configured using the `MLF_HOST`
            environment variable. In case environment variable and argument is passed,
            the URL passed via this argument will take precedence.
        disable_analytics (bool, optional): To turn off usage analytics collection, pass `True`.
            By default, this is set to `False`.
        api_key(deprecated) (Optional[str], optional): API key.
            API key can be found at https://app.truefoundry.com/settings. API key can be
            also configured using the `MLF_API_KEY` environment variable. In case the
            environment variable and argument are passed, the value passed via this argument
            will take precedence.

    Returns:
        MlFoundry: Instance of `MlFoundry` class which represents a `run`.

    Examples:
    ### Get client
    Set the API key using the `MLF_API_KEY` environment variable.
    ```
    export MLF_API_KEY="MY-API_KEY"
    ```

    We can then initialize the client, the API key will be picked up from the
    environment variable.
    ```python
    import mlfoundry

    client = mlfoundry.get_client()
    ```

    ### Get client with API key passed via argument.
    ```python
    import mlfoundry

    API_KEY = "MY-API_KEY"
    client = mlfoundry.get_client(api_key=API_KEY)
    ```
    """
    # TODO (chiragjn): Will potentially need to make MlFoundry (and possibly MlFoundryRun) a Singleton instance.
    #                  Since this sets the tracking URI in global namespace, if someone were to call
    #                  get_client again with different tracking uri, the ongoing run's data will start getting
    #                  pushed to another datastore. Or we should not allow passing in tracking URI and just have
    #                  fixed online and offline clients

    if tracking_uri:
        logger.warning(
            "Passing `tracking_uri` is not supported anymore.\n"
            "Use `mlfoundry login --tracking_uri %s` command or `mlfoundry.login(%s)` function call "
            "to login and set tracking uri",
            tracking_uri,
            tracking_uri,
        )

    if api_key:
        logger.warning(
            "Passing `api_key` is not supported anymore.\n"
            "Use `mlfoundry login` command or `mlfoundry.login()` function call"
        )

    user_id = amplitude.NO_USER

    session = None
    if tracking_uri and tracking_uri.startswith("file:"):
        tracking_uri = os.path.join(tracking_uri, constants.MLRUNS_FOLDER_NAME)
        mlflow.set_tracking_uri(tracking_uri)
    else:
        session = init_session()
        user_id = session.user_info.user_id

    amplitude.init(user_id=user_id, disable_analytics=disable_analytics)
    amplitude.track(
        amplitude.Event.GET_CLIENT,
        # tracking whether user is using file:// or https://
        event_properties={
            "tracking_scheme": urllib.parse.urlparse(tracking_uri).scheme
        },
    )
    return MlFoundry(session=session)


class MlFoundry:
    """MlFoundry."""

    def __init__(self, session: Optional[Session] = None):
        """__init__.

        Args:
            tracking_uri (Optional[str], optional): tracking_uri
        """
        self.mlflow_client = MlflowClient()
        if session:
            self.monitoring_client = MonitoringClient(session=session)

    def _get_or_create_project(self, project_name: str, owner: Optional[str]) -> str:
        """_get_or_create_experiment.

        Args:
            project_name (str): The name of the project.
            owner (Optional[str], optional): Owner of the project. If owner is not passed,
                the current user will be used as owner. If the given owner
                does not have the project, it will be created under
                the current user.

        Returns:
            str: The id of the project.
        """
        experiment_name = project_name
        try:
            experiment = self.mlflow_client.get_experiment_by_name(
                experiment_name, owner_subject_id=owner
            )
            if experiment is not None:
                return experiment.experiment_id
            if not owner:
                logger.info(
                    f"project {experiment_name} does not exist. Creating {experiment_name}."
                )
                return self.mlflow_client.create_experiment(experiment_name)
            else:
                logger.warning(
                    f"project {experiment_name} under owner {owner} does not exist. "
                    "looking for project under current user."
                )
                return self._get_or_create_project(
                    project_name=project_name, owner=None
                )
        except MlflowException as e:
            err_msg = (
                f"Error happened in creating or getting project based on project name: "
                f"{experiment_name}. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

    def get_all_projects(self) -> List[str]:
        """Returns a list of project ids accessible by the current user.

        Returns:
            List[str]: A list of project ids.
        """
        amplitude.track(amplitude.Event.GET_ALL_PROJECTS)
        try:
            experiments = self.mlflow_client.list_experiments(view_type=ViewType.ALL)
        except MlflowException as e:
            err_msg = (
                f"Error happened in fetching project names. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

        projects = []
        for e in experiments:
            # Experiment ID 0 represents default project which we are removing.
            if e.experiment_id != "0":
                projects.append(e.name)

        return projects

    def create_run(
        self,
        project_name: str,
        run_name: Optional[str] = None,
        tags: Optional[Dict[str, Any]] = None,
        owner: Optional[str] = None,
        **kwargs,
    ) -> MlFoundryRun:
        """Initialize a `run`.

        In a machine learning experiment `run` represents a single experiment
        conducted under a project.
        Args:
            project_name (str): The name of the project under which the run will be created.
                project_name should only contain alphanumerics (a-z,A-Z,0-9) or hyphen (-).
                The user must have `ADMIN` or `WRITE` access to this project.
            run_name (Optional[str], optional): The name of the run. If not passed, a randomly
                generated name is assigned to the run. Under a project, all runs should have
                a unique name. If the passed `run_name` is already used under a project, the
                `run_name` will be de-duplicated by adding a suffix.
                run name should only contain alphanumerics (a-z,A-Z,0-9) or hyphen (-).
            tags (Optional[Dict[str, Any]], optional): Optional tags to attach with
                this run. Tags are key-value pairs.
            owner (Optional[str], optional): Owner of the project. If owner is not passed,
                the current user will be used as owner. If the given owner
                does not have the project, it will be created under
                the current user.
            kwargs:

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.

        Examples:
        ### Create a run under current user.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()

        tags = {"model_type": "svm"}
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel", tags=tags
        )

        run.end()
        ```

        ### Creating a run using context manager.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        with client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        ) as run:
            # ...
            # Model training code
            ...
        # `run` will be automatically marked as `FINISHED` or `FAILED`.
        ```

        ### Create a run in a project owned by a different user.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()

        tags = {"model_type": "svm"}
        run = client.create_run(
            project_name="my-classification-project",
            run_name="svm-with-rbf-kernel",
            tags=tags,
            owner="bob",
        )
        run.end()
        ```
        """
        amplitude.track(amplitude.Event.CREATE_RUN)

        if not run_name:
            run_name = coolname.generate_slug(2)
            logger.info(
                f"No run_name given. Using a randomly generated name {run_name}."
                " You can pass your own using the `run_name` argument"
            )
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name!r}"
            )

        experiment_id = self._get_or_create_project(project_name, owner=owner)

        if tags is not None:
            NAMESPACE.validate_namespace_not_used(tags.keys())
        else:
            tags = {}

        tags.update(_get_internal_env_vars_values())
        run = self.mlflow_client.create_run(experiment_id, name=run_name, tags=tags)
        mlf_run_id = run.info.run_id

        mlf_run = MlFoundryRun(experiment_id, mlf_run_id, **kwargs)
        # TODO(Rizwan): Revisit this once run lifecycle is formalised
        mlf_run._add_git_info()
        mlf_run._add_python_mlf_version()
        logger.info(f"Run {run.info.fqn!r} has started.")
        return mlf_run

    def get_run(self, run_id: str) -> MlFoundryRun:
        """Get an existing `run` by the `run_id`.

        Args:
            run_id (str): run_id or fqn of an existing `run`.

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.
        """
        amplitude.track(amplitude.Event.GET_RUN)
        if run_id == "" or (not isinstance(run_id, str)):
            raise MlFoundryException(
                f"run_id must be string type and not empty. "
                f"Got {type(run_id)} type with value {run_id}"
            )
        if "/" in run_id:
            return self.get_run_by_fqn(run_id)

        run = self.mlflow_client.get_run(run_id)
        return MlFoundryRun(
            experiment_id=run.info.experiment_id,
            run_id=run.info.run_id,
        )

    def get_run_by_fqn(self, run_fqn: str) -> MlFoundryRun:
        """Get an existing `run` by `fqn`.

        `fqn` stands for Fully Qualified Name. A run `fqn` has the following pattern:
        owner/project_name/run_name

        If user `bob` has created a run `svm` under the project `cat-classifier`,
        the `fqn` will be `bob/cat-classifier/svm`.

        Args:
            run_fqn (str): `fqn` of an existing run.

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.
        """
        run = self.mlflow_client.get_run_by_fqn(run_fqn)
        return MlFoundryRun(
            experiment_id=run.info.experiment_id,
            run_id=run.info.run_id,
        )

    def get_all_runs(
        self, project_name: str, owner: Optional[str] = None
    ) -> pd.DataFrame:
        """Returns all the run name and id present under a project.

        The user must have `READ` access to the project.
        Args:
            project_name (str): Name of the project.
            owner (Optional[str], optional): Owner of the project. If owner is not passed,
                                   the current user will be used as owner.
        Returns:
            pd.DataFrame: dataframe with two columns- run_id and run_name
        """
        amplitude.track(amplitude.Event.GET_ALL_RUNS)
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name}"
            )
        experiment = self.mlflow_client.get_experiment_by_name(
            project_name, owner_subject_id=owner
        )
        if experiment is None:
            return pd.DataFrame(
                columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
            )

        experiment_id = experiment.experiment_id

        try:
            all_run_infos = self.mlflow_client.list_run_infos(
                experiment_id, run_view_type=ViewType.ALL
            )
        except MlflowException as e:
            err_msg = f"Error happened in while fetching runs for project {project_name}. Error details: {e.message}"
            raise MlFoundryException(err_msg) from e

        runs = []

        for run_info in all_run_infos:
            try:
                run = self.mlflow_client.get_run(run_info.run_id)
                run_name = run.info.name or run.data.tags.get(
                    constants.RUN_NAME_COL_NAME, ""
                )
                runs.append((run_info.run_id, run_name))
            except MlflowException as e:
                logger.warning(
                    f"Could not fetch details of run with run_id {run_info.run_id}. "
                    f"Skipping this one. Error details: {e.message}. "
                )

        return pd.DataFrame(
            runs, columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
        )

    def search_runs(
        self,
        project_name: str,
        filter_string: str = "",
        run_view_type: str = "ACTIVE_ONLY",
        order_by: Sequence[str] = ("attribute.start_time DESC",),
        owner: Optional[str] = None,
    ) -> Generator[MlFoundryRun, None, None]:
        """
        The user must have `READ` access to the project.
        Returns an Generator that returns a MLFoundryRun on each next call.
        All the runs under a project which matches the filter string and the run_view_type are returned.

        Args:
            project_name (str): Name of the project.
            filter_string (str, optional):
                Filter query string, defaults to searching all runs. Identifier required in the LHS of a search expression.
                Signifies an entity to compare against. An identifier has two parts separated by a period: the type of the entity and the name of the entity.
                The type of the entity is metrics, params, attributes, or tags. The entity name can contain alphanumeric characters and special characters.
                You can search using two run attributes : status and artifact_uri. Both attributes have string values.
                When a metric, parameter, or tag name contains a special character like hyphen, space, period, and so on,
                enclose the entity name in double quotes or backticks, params."model-type" or params.`model-type`

            run_view_type (str, optional): one of the following values "ACTIVE_ONLY", "DELETED_ONLY", or "ALL" runs.
            order_by (List[str], optional):
                List of columns to order by (e.g., "metrics.rmse"). Currently supported values
                are metric.key, parameter.key, tag.key, attribute.key. The ``order_by`` column
                can contain an optional ``DESC`` or ``ASC`` value. The default is ``ASC``.
                The default ordering is to sort by ``start_time DESC``.
            owner (Optional[str], optional):
                Owner of the project. If owner is not passed, the current user will be used as owner.

        Examples:
            ```python
            import mlfoundry as mlf

            client = mlf.get_client()
            with client.create_run(project_name="my-project", run_name="run-1") as run1:
                run1.log_metrics(metric_dict={"accuracy": 0.74, "loss": 0.6})
                run1.log_params({"model": "LogisticRegression", "lambda": "0.001"})

            with client.create_run(project_name="my-project", run_name="run-2") as run2:
                run2.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4})
                run2.log_params({"model": "SVM"})

            # Search for the subset of runs with logged accuracy metric greater than 0.75
            filter_string = "metrics.accuracy > 0.75"
            runs = client.search_runs(project_name="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7
            filter_string = "metrics.accuracy > 0.7"
            runs = client.search_runs(project_name="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7 and model="LogisticRegression"
            filter_string = "metrics.accuracy > 0.7 and params.model = 'LogisticRegression'"
            runs = client.search_runs(project_name="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7 and order by accuracy in Descending  order
            filter_string = "metrics.accuracy > 0.7"
            order_by = ["metric.accuracy DESC"]
            runs = client.search_runs(
                project_name="my-project", filter_string=filter_string, order_by=order_by
            )

            ```

        Returns:
            Genarator[MlFoundryRun, None, None]: MLFoundryRuns matching the search query.
        """
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name!r}"
            )
        try:
            run_view_type = ViewType.from_string(run_view_type.lower())
        except Exception as e:
            raise MlFoundryException(e) from e

        try:
            experiment = self.mlflow_client.get_experiment_by_name(
                project_name, owner_subject_id=owner
            )
        except MlflowException as e:
            raise MlFoundryException(e) from e  # user doesnot have READ permission

        if experiment is None:
            owner = f"owner - {owner}" if owner is not None else "current user"
            logger.warning(
                f"Project with name {project_name} does not exist under the {owner}."
            )
            return

        experiment_id = experiment.experiment_id

        page_token = None
        done = False
        while not done:
            all_runs = self.mlflow_client.search_runs(
                experiment_ids=[experiment_id],
                filter_string=filter_string,
                run_view_type=run_view_type,
                max_results=SEARCH_MAX_RESULTS_DEFAULT,
                order_by=order_by,
                page_token=page_token,
            )
            page_token = all_runs.token
            for run in all_runs:
                yield MlFoundryRun(run.info.experiment_id, run.info.run_id)
            done = page_token is None

    @staticmethod
    def get_tracking_uri():
        """get_tracking_uri."""
        return mlflow.tracking.get_tracking_uri()

    def get_model(self, fqn: str) -> ModelVersion:
        """
        Get the model version to download contents or load it in memory

        Args:
            fqn (str): Fully qualified name of the model version.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model(fqn="model:truefoundry/user/my-classification-project/my-sklearn-model:1")

            # load the model into memory
            clf = model_version.load()

            # download the model to disk
            temp = tempfile.TemporaryDirectory()
            download_info = model_version.download(path=temp.name)
            print(download_info)
            ```
        """
        return ModelVersion(fqn)

    def get_artifact(self, fqn: str) -> ArtifactVersion:
        """
        Get the artifact version to download contents

        Args:
            fqn (str): Fully qualified name of the artifact version.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact(fqn="artifact:truefoundry/user/my-classification-project/sklearn-artifact:1")

            # download the artifact to disk
            temp = tempfile.TemporaryDirectory()
            download_info = artifact_version.download(path=temp.name)
            print(download_info)
            ```
        """
        return ArtifactVersion(fqn)

    def log_predictions(
        self, model_version_fqn: str, predictions: List[Union[Prediction, Dict]]
    ):
        """log_predictions.

        Args:
            model_version_fqn (str): fqn of model_version where data needs to be logged
            predictions (List[mlf.Prediction]): List of prediction packets of class mlf.Prediction or dictionary

        example:
            ```python
            import mlfoundry as mlf
            client = mlf.get_client()

            client.log_predictions(
                model_version_fqn = "",
                predictions = [
                    mlf.Prediction(
                        data_id = uuid.uuid4().hex,
                        features = {
                            "feature1": "class1",
                            "feature2": 3.33,
                        },
                        prediction_data = {
                            "value": "pred_class1",
                            "probabilities": {
                                "pred_class1": 0.2,
                                "pred_class2": 0.8
                            },
                            "shap_values": {}
                        },
                        occurred_at = datetime.utcnow(),
                        raw_data = {"data": "any_data"}
                    )
                ]
            )
            ```

        """

        self.monitoring_client.log_predictions(
            model_version_fqn=model_version_fqn, predictions=predictions
        )

    def log_actuals(self, model_version_fqn: str, actuals: List[Union[Actual, Dict]]):
        """log_actuals.

        Args:
            model_version_fqn (str): fqn of model_version where data needs to be logged
            actuals: (List[mlf.Actual]): List of actual packets of class mlf.Actual or a dictionary

        example:
            ```python
            import mlfoundry as mlf
            client = mlf.get_client()
            data_id = uuid.uuid4().hex
            client.log_predictions(
                model_version_fqn = "",
                predictions = [
                    mlf.Prediction(
                        data_id = data_id,
                        features = {
                            "feature1": "class1",
                            "feature2": 3.33,
                        },
                        prediction_data = {
                            "value": "pred_class1",
                            "probabilities": {
                                "pred_class1": 0.2,
                                "pred_class2": 0.8
                            },
                            "shap_values": {}
                        },
                        occurred_at = datetime.utcnow(),
                        raw_data = {"data": "any_data"}
                    )
                ]
            )
            client.log_actuals(
                model_version_fqn = "",
                actuals = [
                    mlf.Actual(
                        data_id = data_id,
                        value = "pred_class2"
                    )
                ]
            )
            ```
        """

        self.monitoring_client.log_actuals(
            model_version_fqn=model_version_fqn, actuals=actuals
        )

    def generate_hash_from_data(
        self, features: Dict, timestamp: Optional[datetime] = None
    ):
        """generate_hash_from_data.

        Args:
            features (Dict): features for which you want to generate a unique hash
            timestamp (Optional[datetime]): Optionally pass a timestamp to generate unique has for features and a timestamp

        example:
            ```python
            import mlfoundry as mlf
            client = mlf.get_client()
            data_id = mlf.generate_hash_from_data(
                features = {
                    "features1": 1.22,
                    "feature2" : "class2"
                }
            )
            ```
        """
        return self.monitoring_client.generate_hash_from_data(
            features=features, timestamp=timestamp
        )

    def get_inference_dataset(
        self,
        model_fqn: str,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        actual_value_required: bool = False,
    ):
        """get_inference_dataset.
        Args:
            model_fqn(str): fqn of model for which inference data is required
            start_time(Optional[datetime]): start_time for "occurred_at" field of the prediction
            end_time(Optional[datetime]): start_time for "occurred_at" field of the prediction
            actual_value_required (Optional[bool]): if true, returns inference data rows with both predictions and actuals logged, default false
        example:
            ```python
            import mlfoundry as mlf
            client = mlf.get_client()
            inference_data = client.get_inference_dataset(model_fqn="")
            ```
        """

        # ToDo (@nikp1172) add better logging, consider edge cases for timezones for start_time/end_time
        if not end_time:
            end_time = datetime.now(tz=timezone.utc)
        if not start_time:
            start_time = end_time - timedelta(days=7)
            logger.info(f"start_time not passed, initializing to {start_time}")
        return self.monitoring_client.get_inference_dataset(
            model_fqn=model_fqn,
            start_time=start_time,
            end_time=end_time,
            actual_value_required=actual_value_required,
        )



from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass(frozen=True)
class Schema:
    feature_column_names: Optional[List[str]] = None
    numerical_feature_column_names: Optional[List[str]] = None
    categorical_feature_column_names: Optional[List[str]] = None
    timestamp_column_name: Optional[str] = None
    prediction_column_name: Optional[str] = None
    actual_column_name: Optional[str] = None
    prediction_probability_column_name: Optional[str] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlfoundry',
 'mlfoundry.artifact',
 'mlfoundry.background',
 'mlfoundry.cli',
 'mlfoundry.cli.commands',
 'mlfoundry.dataset',
 'mlfoundry.dataset.whylogs_types',
 'mlfoundry.frameworks',
 'mlfoundry.integrations',
 'mlfoundry.log_types',
 'mlfoundry.log_types.artifacts',
 'mlfoundry.log_types.image',
 'mlfoundry.metrics',
 'mlfoundry.metrics.v1',
 'mlfoundry.metrics.v2',
 'mlfoundry.monitoring',
 'mlfoundry.monitoring.store',
 'mlfoundry.monitoring.store.repositories',
 'mlfoundry.tracking',
 'mlfoundry.vendor',
 'mlfoundry.vendor.pynvml',
 'mlfoundry.webapp']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.26,<4.0.0',
 'amplitude-tracker>=0.0.7,<0.0.8',
 'boto3>=1.14.1,<2.0.0',
 'click>=8.0.0,<9.0.0',
 'coolname>=1.1.0,<2.0.0',
 'fastparquet>=0.7.2,<0.8.0',
 'filelock>=3.8.0,<4.0.0',
 'importlib-metadata>=4.11.3,<5.0.0',
 'numpy>=1.17.0,<2.0.0',
 'packaging>=21.3,<22.0',
 'pandas>=1.0.0,<2.0.0',
 'psutil>=5.9.0,<6.0.0',
 'pyarrow>=5.0.0,<9.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'scikit-learn>=0.24.2,<2.0.0',
 'scipy>=1.5.4,<2.0.0',
 'tfy-mlflow-client==0.0.21',
 'whylogs>=0.6.15,<0.7.0']

entry_points = \
{'console_scripts': ['mlfoundry = '
                     'mlfoundry.cli.cli_interface:create_mlfoundry_cli']}

setup_kwargs = {
    'name': 'mlfoundry',
    'version': '0.6.0rc2',
    'description': 'Building logger package',
    'long_description': "# MlFoundry\n\n![](https://github.com/MyName/my-project/workflows/Project%20Tests/badge.svg)\n\nThis guide is to give an idea of how you can log metrics, paramaters, predictions, models, dataset using MlFoundry. MlFoundry supports logging above mentioned logs to S3 asynchronously.\n\n## API Workflow:\n\nUsing mlfoundry you ca create multiple projects and each projects can have multiple run.\n\n**Example**\n\n1. Project 1\n   - run 1\n   - run 2\n2. Project 2\n   - run 1\n   - run 2\n   - run 3\n   - run 4\n\nEach run under each project will have a unique run_id.\n\n## Quickstart\n\n### 1. Install mlfoundry\n\n```\n# Install via pip\npip install mlfoundry --extra-index-url https://api.packagr.app/public\n```\n\n### 2. Setup AWS credentials\n\nTo setup AWS credentials MlLogs_foundry's CLI can be used. AWS credentials are required to store the artifacts like model, dataset, whylogs_metrics etc in S3 Bucket.\n\nThe CLI requires AWS credentials(SECRET_ACCESS_KEY, ACCESS_KEY_ID)\n\nAll the credentials can be set by running:\n\n`mlfoundry init`\n\n### 3. Initialise mlfoundry\n\nCreate an project and use the project_name to create a run:\n\n```python\n import mlfoundry as mlf\n\n # create a run\n mlf_run = mlf.create_run(project_name=<project-name>)\n```\n\nIf you want to use a previously created run:\n\n```python\n# printing all the runs available and get the run_id\nshow_all_runs()\n\nrun = get_run(run_id=<run_id>)\n```\n\n### 4. Start logging\n\n**To log a model:**\n\n```python\nrun.log_model(sklearn_model, framework=mlf.ModelFramework.SKLEARN)\n```\n\n**To log parameters:**\n\n```python\nrun.log_params({'learning_rate':0.01,\n                  'n_epochs:10'\n                  })\n```\n\n**To log metrics:**\n\n```python\nrun.log_metrics({'accuracy':87,\n                  'f1_score':0.84,\n                  })\n```\n\n**Log predictions synchronously:**\n\nfeature_df: a pd.DataFrame, the input given to the model to make predictions\n\npredictions: must be a list or pd.Series\n\nTo log predictions synchronously:\n\n```python\nrun.log_predictions(self, feature_df, predictions)\n```\n\nTo log predictions asynchronously:\n\n```python\nresponses = run.log_predictions_async(self, feature_df, predictions)\n\n#### To confirm that the log request completed successfully, await for futures to resolve: This is a blocking call\nimport concurrent.futures as cf\nfor response in cf.as_completed(responses):\n  res = response.result()\n```\n\nUsers can additionaly pass in feature_names argument which is a list of feture names for feature_df.\n\n# To release as Python package\n\nUse Github Releases to create a tag on main and release it. This will trigger the workflow and publish the pip package\n\nTags must be of format `vx.x.x`, example `v0.1.0`\n\n# Development instructions\n\n```\ngit clone https://github.com/truefoundry/mlfoundry.git\ncd mlfoundry\nvirtualenv venv\nsource venv/bin/activate\npip install poetry\npoetry install\npre-commit install\n```\n\n# Run Manual QA\n```\ngit clone https://github.com/truefoundry/mlf-test\ncd mlf-test\nvirtualenv venv\nsource venv/bin/activate\npip install -r requirements.txt\ncd mlf_examples\npython main.py  In-this-step-check-for-error-logs\nmlfoundry ui  In-this-step-go-to-the-listed-url-and-play-around-with-the-ui\n```\n",
    'author': 'Abhishek Choudhary',
    'author_email': 'abhichoudhary06@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/truefoundry/mlfoundry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from mlfoundry.enums import ModelFramework
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.frameworks.base_registry import BaseRegistry
from mlfoundry.frameworks.fastai_registry import FastAIRegistry
from mlfoundry.frameworks.gluon_registry import GluonRegistry
from mlfoundry.frameworks.h2o_registry import H2ORegistry
from mlfoundry.frameworks.keras_registry import KerasRegistry
from mlfoundry.frameworks.lightgbm_registry import LightGBMRegistry
from mlfoundry.frameworks.onnx_registry import ONNXRegistry
from mlfoundry.frameworks.paddle_registry import PaddleRegistry
from mlfoundry.frameworks.pytorch_registry import PyTorchRegistry
from mlfoundry.frameworks.sklearn_registry import SkLearnRegistry
from mlfoundry.frameworks.spacy_registry import SpacyRegistry
from mlfoundry.frameworks.statsmodel_registry import StatsModelsRegistry
from mlfoundry.frameworks.tensorflow_registry import TensorflowRegistry
from mlfoundry.frameworks.xgboost_registry import XGBoostRegistry

MODEL_FRAMEWORK_TO_REGISTRY = {
    ModelFramework.SKLEARN: SkLearnRegistry,
    ModelFramework.TENSORFLOW: TensorflowRegistry,
    ModelFramework.PYTORCH: PyTorchRegistry,
    ModelFramework.KERAS: KerasRegistry,
    ModelFramework.XGBOOST: XGBoostRegistry,
    ModelFramework.LIGHTGBM: LightGBMRegistry,
    ModelFramework.FASTAI: FastAIRegistry,
    ModelFramework.H2O: H2ORegistry,
    ModelFramework.ONNX: ONNXRegistry,
    ModelFramework.SPACY: SpacyRegistry,
    ModelFramework.STATSMODELS: StatsModelsRegistry,
    ModelFramework.GLUON: GluonRegistry,
    ModelFramework.PADDLE: PaddleRegistry,
}


def get_model_registry(
    model_framework: ModelFramework, *args, **kwargs
) -> BaseRegistry:
    if model_framework not in MODEL_FRAMEWORK_TO_REGISTRY:
        raise MlFoundryException(f"{model_framework} is not registerd")
    return MODEL_FRAMEWORK_TO_REGISTRY[model_framework](*args, **kwargs)



import click

from mlfoundry import ModelVersion, get_client


@click.group(help="Download model logged with Mlfoundry")
def download():
    ...


@download.command(short_help="Download a logged model")
@click.option(
    "--fqn",
    required=True,
    type=str,
    help="fqn of the model version",
)
@click.option(
    "--path",
    type=click.Path(file_okay=False, dir_okay=True, exists=False),
    required=True,
    help="path where the model will be downloaded",
)
def model(fqn: str, path: str):
    """
    Download the logged model for a run.\n
    """
    get_client()  # TODO (chiragjn): should give an option for custom tracking uri
    model_version = ModelVersion(fqn=fqn)
    download_path = model_version.download(path=path)
    print(f"Downloaded model files to {download_path}")
